# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for interaction with adb."""
import enum
import json
import os
import re
import subprocess
import time
from typing import Any, List, Optional, Tuple

from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils
from gazoo_device.utility import retry

ADB_RETRY_SLEEP = 10
DEFAULT_PORT = 5555
FASTBOOT_TIMEOUT = 10.0
PROPERTY_PATTERN = r"\[(.*)\]: \[(.*)\]\n"
SYSENV_PATTERN = r"(.*)=(.*)\n"


class AdbDeviceState(enum.Enum):
  """ADB device states as found in output of 'adb devices'.

  The values come from connection_state_name() in
  https://android.googlesource.com/platform/system/adb/+/refs/heads/master/transport.cpp#759.
  """
  BOOTLOADER = "bootloader"
  DEVICE = "device"
  HOST = "host"
  OFFLINE = "offline"
  NO_PERMISSIONS = "no permissions"
  RECOVERY = "recovery"
  SIDELOAD = "sideload"
  UNAUTHORIZED = "unauthorized"
  UNKNOWN = "unknown"
  # Not defined by ADB (in case we fail to parse the state).
  UNRECOGNIZED = "unrecognized"


logger = gdm_logger.get_logger()


def bugreport(adb_identifier: str,
              destination_path: str = "./",
              adb_path: Optional[str] = None) -> str:
  """Gets a bugreport to destination_path on host for the adb_identifier.

  Args:
      adb_identifier: Device ADB identifier, a serial number ("abcde123") or an
          IP address and a port number ("12.34.56.78:5555").
      destination_path: Path to destination on host computer where bugreport
          should copied to.
      adb_path: optional alternative path to adb executable. If adb_path is not
          provided then path returned by get_adb_path will be used instead.

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
          get_adb_path or bugreport failed.
      ValueError: if destination_path does not exist.

  Returns:
      Output from calling 'adb bugreport'.
  """
  destination_dir = os.path.dirname(destination_path)
  if not os.path.exists(destination_dir):
    raise ValueError(f"The destination_path directory {destination_dir} does "
                     "not exist.")
  output, returncode = _adb_command(command=["bugreport", destination_path],
                                    adb_serial=adb_identifier,
                                    adb_path=adb_path,
                                    include_return_code=True)
  if returncode != 0:
    raise RuntimeError(f"Getting bugreport on ADB device {adb_identifier} to "
                       f"{destination_path} failed. "
                       f"Error: {output!r} with return code {returncode}")
  return output


def enter_fastboot(adb_serial, adb_path=None):
  """Enters fastboot mode by calling 'adb reboot bootloader' for the adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb reboot' or None if call fails with
      non-zero
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command(("reboot", "bootloader"), adb_serial, adb_path=adb_path)


def exit_fastboot(fastboot_serial,
                  fastboot_path=None,
                  timeout=FASTBOOT_TIMEOUT):
  """Exits fastboot mode by calling 'fastboot reboot' for the fastboot_serial provided.

  Args:
      fastboot_serial (str): Device fastboot serial number.
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot reboot to return

  Raises:
      RuntimeError: if fastboot_path is invalid or fastboot executable was not
                    found by get_fastboot_path.

  Returns:
      str: Output from calling 'fastboot reboot' or None if call fails with
      non-zero
           return code.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead. If fastboot returns a non-zero return code then
      None will be returned.
  """
  if fastboot_path is None:
    fastboot_path = get_fastboot_path()

  if not os.path.exists(fastboot_path):
    raise RuntimeError(
        "The fastboot_path of {} appears to be invalid.".format(fastboot_path))

  try:
    args = ("timeout", str(timeout), fastboot_path, "-s", fastboot_serial,
            "reboot")
    return subprocess.check_output(
        args, stderr=subprocess.STDOUT).decode("utf-8", "replace")
  except subprocess.CalledProcessError:
    return None


def fastboot_unlock_device(fastboot_serial,
                           fastboot_path=None,
                           timeout=FASTBOOT_TIMEOUT):
  """Unlock the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(("flashing", "unlock"),
                           fastboot_serial=fastboot_serial,
                           fastboot_path=fastboot_path,
                           timeout=timeout)


def fastboot_lock_device(fastboot_serial,
                         fastboot_path=None,
                         timeout=FASTBOOT_TIMEOUT):
  """Lock the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(("flashing", "lock"),
                           fastboot_serial=fastboot_serial,
                           fastboot_path=fastboot_path,
                           timeout=timeout)


def fastboot_check_is_unlocked(fastboot_serial: str,
                               fastboot_path: Optional[str] = None,
                               timeout: float = FASTBOOT_TIMEOUT) -> bool:
  """Checks if the device is unlocked.

  Args:
      fastboot_serial: Device serial number
      fastboot_path: Optional alternative path to fastboot executable
      timeout: In seconds to wait for fastboot command to return

  Raises:
      RuntimeError: If the response is not in the expected format.

  Returns:
      True if the device is unlocked, else False.
  """
  raw_result = _fastboot_command(("getvar", "unlocked"),
                                 fastboot_serial=fastboot_serial,
                                 fastboot_path=fastboot_path,
                                 timeout=timeout)
  # Raw result would be like "unlocked: yes"
  match = re.search(r"unlocked: (yes|no)", raw_result)
  if not match:
    raise RuntimeError(
        f"Unknown output from fastboot getvar unlocked: {raw_result}")
  return match.group(1) == "yes"


def fastboot_wipe_userdata(fastboot_serial,
                           fastboot_path=None,
                           timeout=FASTBOOT_TIMEOUT):
  """Wipe user data on the device through fastboot.

  Args:
      fastboot_serial (str): Device serial number
      fastboot_path (str): optional alternative path to fastboot executable
      timeout (float): in seconds to wait for fastboot command to return

  Returns:
      str: response from fastboot command
  """
  return _fastboot_command(
      "-w",
      fastboot_serial=fastboot_serial,
      fastboot_path=fastboot_path,
      timeout=timeout)


def enter_sideload(adb_serial, adb_path=None, auto_reboot=False):
  """Enters sideload mode by calling 'adb reboot sideload' for the adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable.
      auto_reboot (bool): whether to auto reboot after sideload complete.

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from command call.
  """
  if auto_reboot:
    command = ("reboot", "sideload-auto-reboot")
  else:
    command = ("reboot", "sideload")
  return _adb_command(command, adb_serial=adb_serial, adb_path=adb_path)


def sideload_package(package_path, adb_serial, adb_path=None):
  """Perform "adb sideload <package>" command.

  Args:
      package_path (str): the path of the package to sideload.
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable.

  Returns:
      str: the command output.

  Raises:
      RuntimeError: if package_path is invalid.
  """
  if not os.path.isfile(package_path):
    raise RuntimeError(
        "sideload_package failed: {} is not a file.".format(package_path))
  return _adb_command(("sideload", package_path),
                      adb_serial=adb_serial,
                      adb_path=adb_path)


def adb_devices(
    adb_path: Optional[str] = None,
    state: Optional[AdbDeviceState] = None
) -> List[Tuple[str, AdbDeviceState]]:
  """Returns parsed output of 'adb devices'.

  Args:
    adb_path: Optional alternative path to the 'adb' executable. Defaults to the
      path returned by get_adb_path().
    state: If provided, only include devices in the given ADB state.

  Returns:
    List of (device identifier, device ADB state) tuples found in
    'adb devices'. The device identifiers are either serial numbers
    ("abcde123") or IP addresses and ports ("12.34.56.78:5555").
  """
  try:
    output = _adb_command("devices", adb_path=adb_path)
  except RuntimeError as err:
    logger.warning(repr(err))
    return []

  output_lines = output.splitlines()
  output_start_marker = "List of devices attached"
  if output_start_marker not in output_lines:
    return []

  output_start_index = output_lines.index(output_start_marker) + 1
  device_lines = [line for line in output_lines[output_start_index:]
                  if line and "\t" in line]

  identifiers_and_states = []
  for device_line in device_lines:
    identifier, _, device_state_str = device_line.partition("\t")
    if "no permissions" in device_state_str:
      # Reasons and URLs vary: "no permissions (<reason>); see [<url>]".
      device_state = AdbDeviceState.NO_PERMISSIONS
    else:
      try:
        device_state = AdbDeviceState(device_state_str)
      except ValueError as e:
        logger.debug(
            f"Failed to parse ADB state {device_state_str!r}. Error: {e!r}")
        device_state = AdbDeviceState.UNRECOGNIZED
    if state is None or state == device_state:
      identifiers_and_states.append((identifier, device_state))
  return identifiers_and_states


def get_adb_devices(adb_path: Optional[str] = None) -> List[str]:
  """Returns ADB device identifiers of available ("device") ADB devices.

  Args:
    adb_path: Optional alternative path to the 'adb' executable. Defaults to the
      path returned by get_adb_path().

  Returns:
    Returns ADB device identifiers of available ("device") ADB devices. The
    device identifiers are either serial numbers ("abcde123") or IP addresses
    and ports ("12.34.56.78:5555").
  """
  return [identifier
          for identifier, _ in adb_devices(adb_path, AdbDeviceState.DEVICE)]


def get_sideload_devices(adb_path: Optional[str] = None) -> List[str]:
  """Returns ADB device identifiers of devices in "sideload" mode.

  Args:
    adb_path: Optional alternative path to the 'adb' executable. Defaults to
        the path returned by get_adb_path().

  Returns:
    Returns ADB device identifiers of devices in "sideload" mode. The device
    identifiers are either serial numbers ("abcde123") or IP addresses and ports
    ("12.34.56.78:5555").
  """
  return [identifier
          for identifier, _ in adb_devices(adb_path, AdbDeviceState.SIDELOAD)]


def is_adb_mode(adb_identifier: str, adb_path: Optional[str] = None) -> bool:
  """Returns whether the ADB identifier is shown as a "device" in 'adb devices'.

  Args:
    adb_identifier: Device ADB identifier, a serial number ("abcde123") or an IP
      address and a port number ("12.34.56.78:5555").
    adb_path: An optional alternative path to the 'adb' executable. If not
        provided, the path returned by get_adb_path() is used instead.
  """
  is_available = adb_identifier in get_adb_devices(adb_path=adb_path)
  if not re.search(host_utils.IP_ADDRESS, adb_identifier):
    return is_available
  ip_address = adb_identifier.split(":")[0]  # Remove the port number.
  # Devices connected through ADB over IP can show up as available ("device") in
  # 'adb devices' even if the device is offline.
  return is_available and host_utils.is_pingable(ip_address)


def is_sideload_mode(
    adb_identifier: str, adb_path: Optional[str] = None) -> bool:
  """Returns True if the adb_identifier is in sideload mode.

  Args:
    adb_identifier: Device ADB identifier, a serial number ("abcde123") or an IP
      address and a port number ("12.34.56.78:5555").
    adb_path: An optional alternative path to the 'adb' executable. If not
        provided, the path returned by get_adb_path() is used instead.
  """
  return adb_identifier in get_sideload_devices(adb_path=adb_path)


def get_adb_path(adb_path=None):
  """Returns the correct adb path to use.

  Args:
      adb_path (str): path to "adb" executable.

  Notes: Starts with passed in path, then looks at config, and finally
    system's default adb if available.

  Raises:
      RuntimeError: if no valid adb path could be found

  Returns:
      str: Path to correct adb executable to use.
  """
  if is_valid_path(adb_path):
    return adb_path
  try:
    with open(config.DEFAULT_GDM_CONFIG_FILE) as config_file:
      gdm_config = json.load(config_file)
    adb_path = gdm_config[config.ADB_BIN_PATH_CONFIG]
  except (IOError, KeyError, ValueError):
    pass

  if is_valid_path(adb_path):
    return adb_path
  elif adb_path:
    logger.warning("adb path {!r} stored in {} does not exist."
                   .format(adb_path, config.DEFAULT_GDM_CONFIG_FILE))

  if host_utils.has_command("adb"):
    return host_utils.get_command_path("adb")
  raise RuntimeError("No valid adb path found using 'which adb'")


def is_valid_path(path):
  return path and os.path.exists(path)


def get_adb_over_ip_identifier(
    adb_identifier: str, port: int = DEFAULT_PORT) -> str:
  """Returns an ADB over IP identifier in the format 'IP_address:port'.

  If the provided adb_identifier does not contain a port number, the default
  port (":5555") is appended to the identifier.

  Args:
    adb_identifier: IP address or IP address and a port number.
    port: TCP port number.

  Raises:
    ValueError: if adb_identifier is not an IP address.
  """
  if not re.search(host_utils.IP_ADDRESS, adb_identifier):
    raise ValueError(f"ADB identifier {adb_identifier!r} is not an IP address.")
  if ":" not in adb_identifier:
    adb_identifier = f"{adb_identifier}:{port}"
  return adb_identifier


def _is_connect_successful(output_and_return_code: Tuple[str, int]) -> bool:
  """Returns whether the connect() call was successful.

  Args:
    output_and_return_code: Output and return code of 'adb connect' command.

  Returns:
    True if the return code of 'adb connect' is 0 and there are no failure
    markers in the command output. Note that 'adb connect' can return a 0 exit
    code even when it fails, which is why we have to check for failure markers.
  """
  output, return_code = output_and_return_code
  failure_markers = [
      "failed to connect",
      "failed to resolve host",
      "missing port in specification",
      "unable to connect",
  ]
  return (return_code == 0 and
          all(marker not in output for marker in failure_markers))


def connect(adb_identifier: str, attempts: int = 3) -> str:
  """Connects to the device via ADB and returns the command output.

  Args:
    adb_identifier: IP address ("12.34.56.78") or IP address and a port number
      ("12.34.56.78:5555"). If a port number is not provided, defaults to 5555.
    attempts: Number of attempts for performing 'adb connect'.

  Raises:
    DeviceError: if 'adb connect' fails, or adb_identifier is not found in
      'adb devices' after 'adb connect'.

  Returns:
    Output of the 'adb connect' command.
  """
  adb_identifier = get_adb_over_ip_identifier(adb_identifier)

  retry_interval_s = 3
  try:
    output, _ = retry.retry(
        func=_adb_command,
        func_kwargs={
            "command": ["connect", adb_identifier],
            "include_return_code": True,
        },
        is_successful=_is_connect_successful,
        timeout=retry_interval_s * attempts,
        interval=retry_interval_s,
        reraise=False)
  except errors.CommunicationTimeoutError as e:
    raise errors.DeviceError(
        f"Unable to connect to device {adb_identifier!r} via ADB. Error: {e!r}")

  try:
    retry.retry(
        func=is_adb_mode,
        func_args=[adb_identifier],
        is_successful=bool,
        timeout=retry_interval_s * attempts,
        interval=retry_interval_s)
  except errors.CommunicationTimeoutError as e:
    raise errors.DeviceError(
        f"{adb_identifier!r} was not found in 'adb devices' after "
        f"'adb connect'. Error: {e!r}")
  return output


def _is_disconnected(adb_identifier: str) -> bool:
  """Returns True if adb_identifier is not present in 'adb devices'."""
  return adb_identifier not in [adb_id for adb_id, _ in adb_devices()]


def disconnect(adb_identifier: str, attempts: int = 3) -> str:
  """Disconnects ADB from the device and returns the command output.

  Args:
    adb_identifier: IP address ("12.34.56.78") or IP address and a port number
      ("12.34.56.78:5555"). If a port number is not provided, defaults to 5555.
    attempts: Number of attempts for performing 'adb disconnect'.

  Raises:
    DeviceError: if 'adb disconnect' fails, or if adb_identifier is still
      present in 'adb devices' after 'adb disconnect'.

  Returns:
    Output of the 'adb disconnect' command.
  """
  adb_identifier = get_adb_over_ip_identifier(adb_identifier)

  retry_interval_s = 3
  try:
    output, _ = retry.retry(
        func=_adb_command,
        func_kwargs={
            "command": ["disconnect", adb_identifier],
            "include_return_code": True,
        },
        is_successful=lambda output_and_code: output_and_code[1] == 0,
        timeout=retry_interval_s * attempts,
        interval=retry_interval_s,
        reraise=False)
  except errors.CommunicationTimeoutError as e:
    raise errors.DeviceError(
        f"Unable to disconnect ADB from device {adb_identifier!r}. "
        f"Error: {e!r}")

  try:
    retry.retry(
        func=_is_disconnected,
        func_kwargs={"adb_identifier": adb_identifier},
        is_successful=bool,
        timeout=retry_interval_s * attempts,
        interval=retry_interval_s)
  except errors.CommunicationTimeoutError as e:
    raise errors.DeviceError(
        f"{adb_identifier!r} was still found in 'adb devices' after "
        f"'adb connect'. Error: {e!r}")
  return output


def shell(adb_serial: str,
          command: str,
          adb_path: Optional[str] = None,
          timeout: Optional[int] = None,
          retries: int = 1,
          include_return_code: bool = False) -> Any:
  """Issues a command to the shell of the adb_serial provided.

  Args:
    adb_serial: Device serial number.
    command: Command to send.
    adb_path: Optional alternative path to adb executable.
    timeout: Time in seconds to wait for adb process to complete.
    retries: Number of times to retry adb command.
    include_return_code: A flag indicating return code should also be
      returned.

  Returns:
    Response string if include_return_code is False; (response, return code)
      tuple otherwise.
  """
  return _adb_command(["shell", command], adb_serial,
                      adb_path=adb_path, timeout=timeout, retries=retries,
                      include_return_code=include_return_code)


def get_fastboot_devices(fastboot_path=None):
  """Returns list of ADB devices in fastboot (bootloader) mode.

  Args:
      fastboot_path (str): optional alternative path to fastboot executable

  Returns:
      list: A list of ADB device serial numbers in fastboot mode.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead.
      If fastboot path invalid, will return empty list.
  """
  try:
    fastboot_path = get_fastboot_path(fastboot_path)
  except RuntimeError as err:
    logger.warning(repr(err))
    return []

  try:
    output = subprocess.check_output((fastboot_path, "devices"),
                                     stderr=subprocess.STDOUT)
    output = output.decode("utf-8", "replace")
    device_lines = [x for x in output.splitlines() if "astboot" in x]
    return [x.split()[0] for x in device_lines]
  except subprocess.CalledProcessError:
    return []


def get_fastboot_path(fastboot_path=None):
  """Returns the fastboot executable path to use.

  Args:
      fastboot_path (str): path to "fastboot" executable.

  Raises:
      RuntimeError: if no valid fastboot executable could be found

  Returns:
      str: Path to correct fastboot executable to use.
  """
  if is_valid_path(fastboot_path):
    return fastboot_path
  if host_utils.has_command("fastboot"):
    return host_utils.get_command_path("fastboot")
  raise RuntimeError("No valid fastboot path found using 'which fastboot'")


def is_device_online(adb_serial, adb_path=None, fastboot_path=None):
  """Returns true if the device appears in either 'adb devices' or 'fastboot devices'.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable
      fastboot_path (str): optional alternative path to fastboot executable

  Returns:
      bool: True if device is in adb or fastboot mode. False otherwise.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If fastboot_path is not provided then path returned by
      get_fastboot_path will be used instead.
  """
  return (is_adb_mode(adb_serial, adb_path=adb_path) or
          is_fastboot_mode(adb_serial, fastboot_path=fastboot_path))


def is_fastboot_mode(adb_serial, fastboot_path=None):
  """Checks if device is in fastboot mode.

  Args:
      adb_serial (str): Device serial number.
      fastboot_path (str): optional alternative path to fastboot executable

  Raises:
      RuntimeError: if fastboot_path is invalid or fastboot executable was
                    not found by get_fastboot_path.

  Returns:
      bool: True if device is in fastboot mode. False otherwise.

  Note:
      If fastboot_path is not provided then path returned by get_fastboot_path
      will be used instead.
  """
  return adb_serial in get_fastboot_devices(fastboot_path=fastboot_path)


def pull_from_device(adb_serial, sources, destination_path="./", adb_path=None):
  """Pulls sources from device to destination_path on host for adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      sources (str or list): Path to one or more source files on device to
        copy to host.
      destination_path (str): Path to destination on host computer where file
        should copied to.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path or push failed.
      ValueError: if destination_path directory doesn't exist.

  Returns:
      str: Output from calling 'adb push' or None if call raises an erroro
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned. If no destination_path is provided the file will be copied to
      the current working directory on the host computer.
  """
  destination_dir = os.path.dirname(destination_path)
  if destination_dir != "." and not os.path.exists(destination_dir):
    raise ValueError(
        "The destination_path directory {} appears to be invalid.".format(
            destination_dir))

  args = ["pull"]
  if isinstance(sources, list):
    for source_path in sources:
      args.append(source_path)
  else:
    args.append(sources)
  args.append(destination_path)
  output, returncode = _adb_command(
      args, adb_serial, adb_path=adb_path, include_return_code=True)
  if returncode != 0:
    raise RuntimeError("Pulling file(s) {} on ADB device {} to {} failed. "
                       "Error: {!r}".format(sources, adb_serial,
                                            destination_path, output))
  return output


def push_to_device(adb_serial, sources, destination_path, adb_path=None):
  """Pushes sources to destination_path on device for adb_serial provided.

  Args:
      adb_serial (str): Device serial number.
      sources (str or list): Path to one or more source files on host computer
        to copy to device.
      destination_path (str): Path to destination on device where file should
        copied to.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path or push failed.
      ValueError: if source_path doesn't exist.

  Returns:
      str: Output from calling 'adb push' or None if call raises an erroro
           return code.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  args = ["push"]
  if isinstance(sources, list):
    for source_path in sources:
      args.append(source_path)
      if not os.path.exists(source_path):
        raise ValueError(
            "The source file {} appears to be invalid.".format(source_path))
  else:
    args.append(sources)
    if not os.path.exists(sources):
      raise ValueError(
          "The source file {} appears to be invalid.".format(sources))

  args.append(destination_path)
  output, returncode = _adb_command(
      args, adb_serial, adb_path=adb_path, include_return_code=True)
  if returncode != 0:
    raise RuntimeError("Pushing file(s) {} to {} on ADB device {} failed. "
                       "Error: {!r}".format(sources, destination_path,
                                            adb_serial, output))
  return output


def reboot_device(adb_serial, adb_path=None, retries=1):
  """Calls 'adb reboot' for the adb_serial provided using adb_path.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executables
      retries (int): number of times to retry adb command.

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb reboot'.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command("reboot", adb_serial, adb_path=adb_path, retries=retries)


def root_device(adb_serial, adb_path=None):
  """Calls 'adb root' for the adb_serial provided using adb_path.

  Args:
      adb_serial (str): Device serial number.
      adb_path (str): optional alternative path to adb executable

  Raises:
      RuntimeError: if adb_path is invalid or adb executable was not found by
                    get_adb_path.

  Returns:
      str: Output from calling 'adb root'.

  Note:
      If adb_path is not provided then path returned by get_adb_path will be
      used instead. If adb returns a non-zero return code then None will be
      returned.
  """
  return _adb_command("root", adb_serial, adb_path=adb_path)


def verify_user_has_fastboot(device_name):
  """Verifies fastboot available and user is root or in plugdev group.

  Args:
      device_name (str): Device name to use in error output.

  Raises:
      DeviceError: Fastboot is not on computer OR
                   'plugdev' group doesn't exist OR
                   current user is not in the 'plugdev' group.
  """
  if not host_utils.has_command("fastboot"):
    raise errors.DeviceError("Device {} verify user has fastboot failed. "
                             "Fastboot executable is not installed. "
                             "See readme about installing adb (which installs "
                             "fastboot) then su -$USER (or logout and back in) "
                             "to add user to plugdev group".format(device_name))


def wait_for_device(adb_identifier: str, timeout: float) -> None:
  """Waits until the device is detected by ADB.

  Args:
      adb_identifier: ADB device identifier, e.g. serial number or IP address.
      timeout: Time in seconds to wait before giving up.

  Raises:
      CommunicationTimeoutError: Timed out waiting for device detection.
  """
  _, return_code = _adb_command(command="wait-for-device",
                                adb_serial=adb_identifier,
                                timeout=timeout,
                                include_return_code=True)
  if return_code != 0:
    raise errors.CommunicationTimeoutError(
        f"Timeout waiting for device {adb_identifier} after {timeout} seconds")


def wait_for_device_offline(adb_identifier: str, timeout: float = 20,
                            check_interval: float = 1) -> None:
  """Waits until the device is not seen via adb, e.g. reboot started.

  Args:
      adb_identifier: ADB device identifier, e.g. serial number or IP address.
      timeout: Time in seconds to wait before raising timeout.
      check_interval: Interval between checks to see if device is offline.

  Raises:
      CommunicationTimeoutError: Timed out waiting for device to go offline.
  """
  end_time = time.time() + timeout
  while is_adb_mode(adb_identifier):
    if time.time() > end_time:
      raise errors.CommunicationTimeoutError(
          f"Device {adb_identifier} did not go offline in {timeout} seconds.")
    time.sleep(check_interval)


def _adb_command(command,
                 adb_serial=None,
                 adb_path=None,
                 include_return_code=False,
                 timeout=None,
                 retries=1):
  """Returns the output of the adb command and optionally the return code.

  Args:
      command (str or tuple): ADB command and optionally arguments to execute.
      adb_serial (str): Device serial number
      adb_path (str): optional alternative path to adb executable
      include_return_code (bool): flag indicating return code should also be
        returned.
      timeout (int): time in seconds to wait for adb process to complete.
      retries (int): number of times to retry adb command.

  Raises:
      RuntimeError: if adb_path provided or obtained from get_adb_path is
                    invalid (executable at path doesn't exist).

  Returns:
      str: The ADB command output (including stderr)
      tuple: The ADB command output (including stderr) and return code

  Note:
      The stderr is redirected to stdout so callers should use the return code
      or search the output for known errors if they want to determine if the
      command succeeded or not.
  """
  adb_path = get_adb_path(adb_path)

  if adb_serial is None:
    args = [adb_path]
  else:
    args = [adb_path, "-s", adb_serial]
  if isinstance(command, str):
    args.append(command)
  elif isinstance(command, (list, tuple)):
    args.extend(command)
  for i in range(0, retries):
    proc = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
      output, _ = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
      proc.terminate()
      output, _ = proc.communicate()
    output = output.decode("utf-8", "replace")
    logger.debug("adb command {!r} to {} returned {!r}".format(
        command, adb_serial, output))
    if include_return_code:
      return output, proc.returncode
    adb_failure_messages = ["error: closed", "adb: device offline"]
    if not any(msg in output for msg in adb_failure_messages):
      return output
    if i < retries - 1:
      logger.info(f"Retrying adb command: {command} in {ADB_RETRY_SLEEP}s")
      time.sleep(ADB_RETRY_SLEEP)
  raise errors.DeviceError(
      f"ADB command failed: {command} with output: {output}")


def _fastboot_command(command,
                      fastboot_serial=None,
                      fastboot_path=None,
                      include_return_code=False,
                      timeout=FASTBOOT_TIMEOUT):
  """Returns the output of the fastboot command and optionally the return code.

  Args:
      command (str or tuple): fastboot command and optionally arguments to
        execute.
      fastboot_serial (str): Device fastboot serial number.
      fastboot_path (str): optional alternative path to fastboot executable
      include_return_code (bool): flag indicating return code should also be
        returned.
      timeout (float): in seconds to wait for fastboot command to return

  Raises:
      RuntimeError: if fastboot_path provided or obtained from
      get_fastboot_path is invalid (executable at path doesn't exist).

  Returns:
      str: The fastboot command output (including stderr)
      tuple: The fastboot command output (including stderr) and return code

  Note:
      The stderr is redirected to stdout so callers should use the return code
      or search the output for known errors if they want to determine if the
      command succeeded or not.
  """
  if fastboot_path is None:
    fastboot_path = get_fastboot_path()
  if not os.path.exists(fastboot_path):
    raise RuntimeError(
        "The fastboot_path of {} appears to be invalid.".format(fastboot_path))

  if fastboot_serial is None:
    args = ["timeout", str(timeout), fastboot_path]
  else:
    args = ["timeout", str(timeout), fastboot_path, "-s", fastboot_serial]

  if isinstance(command, str):
    args.append(command)
  elif isinstance(command, (list, tuple)):
    args.extend(command)

  proc = subprocess.Popen(
      args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  output, _ = proc.communicate()
  output = output.decode("utf-8", "replace")
  if include_return_code:
    return output, proc.returncode
  return output


def install_package_on_device(package_path: str,
                              adb_serial: Optional[str] = None,
                              adb_path: Optional[str] = None,
                              allow_downgrade: bool = False,
                              allow_test_apk: bool = False,
                              reinstall: bool = False,
                              all_permissions: bool = False) -> None:
  """Installs an apk on a target device.

  Use adb install command to install a package to the system.
  The options are subjected to the adb install command. See the doc.
  https://developer.android.com/studio/command-line/adb#shellcommands

  Args:
      package_path: The path to the package on host machine.
      adb_serial: The device serial, optional.
      adb_path: An optional alternative path to adb executable.
      allow_downgrade: Allows version code downgrade.
      allow_test_apk: Allows test APKs to be installed.
      reinstall: Reinstalls an existing app and keeps its data.
      all_permissions: Grants all runtime permission to the app.

  Raises:
      ValueError: when package_path is not valid.
      DeviceError: when installation failed.
  """
  if not os.path.exists(package_path):
    raise ValueError(
        "install_package_on_device received invalid package_path: {}".format(
            package_path))

  flags_map = {
      "-d": allow_downgrade,
      "-t": allow_test_apk,
      "-r": reinstall,
      "-g": all_permissions,
  }
  command_list = ["install"]
  flags = sorted([flag for flag, value in flags_map.items() if value])
  command_list.extend(flags)
  command_list.append(package_path)
  response = _adb_command(
      tuple(command_list), adb_serial=adb_serial, adb_path=adb_path)
  if "Success\n" not in response:
    raise errors.DeviceError(
        "install_package_on_device failed: {}".format(response))


def uninstall_package_on_device(package_name, adb_serial=None, adb_path=None):
  """Uninstall a package on a target device.

  Args:
      package_name (str): the name of the package, e.g.,
        "com.google.android.apps.somepackage.someapp".
      adb_serial (str): the device serial, optional.
      adb_path (str): optional alternative path to adb executable.

  Raises:
      DeviceError: when uninstall failed.
  """
  response = _adb_command(("uninstall", package_name),
                          adb_serial=adb_serial,
                          adb_path=adb_path)
  if "Success\n" not in response:
    raise errors.DeviceError("uninstall_package_on_device failed.")


def add_port_forwarding(host_port: int,
                        device_port: int,
                        adb_serial: Optional[str] = None,
                        adb_path: Optional[str] = None) -> str:
  """Forwards the socket connection from host to device.

  Args:
      host_port: The port on the host side.
      device_port: The port on the device side.
      adb_serial: The device serial, optional.
      adb_path: Alternative path to adb executable.

  Raises:
      RuntimeError: If the port forwarding doesn't exist.

  Returns:
      The command output.
  """
  commands = ("forward", f"tcp:{host_port}", f"tcp:{device_port}")
  output, returncode = _adb_command(commands,
                                    adb_serial=adb_serial,
                                    adb_path=adb_path,
                                    include_return_code=True)
  if returncode != 0:
    raise RuntimeError(
        f"Failed to add port forwarding on device serial:{adb_serial} from "
        f"host port {host_port} to device port {device_port}. {output}")
  return output


def remove_port_forwarding(host_port: int,
                           adb_serial: Optional[str] = None,
                           adb_path: Optional[str] = None) -> str:
  """Removes a forward socket connection.

  Args:
      host_port: The port on the host side.
      adb_serial: The device serial, optional.
      adb_path: Alternative path to adb executable.

  Raises:
      RuntimeError: If the port forwarding doesn't exist.

  Returns:
      The command output.
  """
  commands = ("forward", "--remove", f"tcp:{host_port}")
  output, returncode = _adb_command(commands,
                                    adb_serial=adb_serial,
                                    adb_path=adb_path,
                                    include_return_code=True)
  if returncode != 0:
    raise RuntimeError(
        f"Failed to remove port forwarding on device serial: {adb_serial} "
        f"on host port {host_port}. {output}")
  return output


def tcpip(adb_serial: Optional[str] = None,
          port: Optional[int] = DEFAULT_PORT,
          adb_path: Optional[str] = None) -> str:
  """Restarts adbd listening on the provided TCP port for the adb_serial.

  Args:
    adb_serial: ADB serial number.
    port: TCP port to listen on.
    adb_path: Alternative path to 'adb' executable.

  Raises:
    RuntimeError: If the command fails.

  Returns:
    The command output.
  """
  output, return_code = _adb_command(
      ["tcpip", str(port)],
      adb_serial=adb_serial,
      adb_path=adb_path,
      include_return_code=True)
  if return_code != 0:
    raise RuntimeError(
        f"ADB failed to start listening on port {port} for ADB serial "
        f"{adb_serial}. Return code: {return_code}, output: {output}.")
  return output



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package management capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class PackageManagementBase(capability_base.CapabilityBase):
  """Abstract base class defining API for package management."""

  @abc.abstractmethod
  def install_package(self, package_path: str) -> None:
    """Installs a package on the device.

    Args:
      package_path: The path to the package on the host machine.

    Raises:
      ValueError: when package_path is not valid.
      DeviceError: if failed to install the package.
    """

  @abc.abstractmethod
  def uninstall_package(self, package_name: str) -> None:
    """Uninstalls a package on the device.

    Args:
      package_name: The name of the package to uninstall.

    Raises:
      DeviceError: if failed to uninstall the package.
    """

  @abc.abstractmethod
  def has_package(self, package_name: str) -> bool:
    """Returns whether the package is installed on the device.

    Args:
      package_name: The package name.
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of the Matter Level Control cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import level_control_base
from gazoo_device.protos import attributes_service_pb2

logger = gdm_logger.get_logger()
LevelControlCluster = matter_enums.LevelControlCluster
INT8U_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT8U_ATTRIBUTE_TYPE


class LevelControlClusterPwRpc(level_control_base.LevelControlClusterBase):
  """Matter Level Control cluster capability."""

  @decorators.CapabilityLogDecorator(logger)
  def move_to_level(self, level: int, verify: bool = True) -> None:
    """The MoveToLevel command.

    On receipt of this command, a device should move from its current level
    to the value given in the Level field.
    The meaning of "level" is device dependent. (For a light it may mean
    brightness level)

    Args:
      level: The level that the device should move to.
      verify: If true, verifies the level changes before returning.
    """
    previous_level = self.current_level

    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=LevelControlCluster.ID,
        attribute_id=LevelControlCluster.ATTRIBUTE_CURRENT_LEVEL,
        attribute_type=INT8U_ATTRIBUTE_TYPE,
        data_uint8=level)

    if verify:
      if self.current_level != level:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} current level didn't change to "
            f"{level} from {previous_level}.")

  @decorators.DynamicProperty
  def current_level(self) -> int:
    """The CurrentLevel attribute.

    The CurrentLevel attribute represents the current level of this device.
    The meaning of "level" is device dependent.

    Returns:
      The current level.
    """
    level_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=LevelControlCluster.ID,
        attribute_id=LevelControlCluster.ATTRIBUTE_CURRENT_LEVEL,
        attribute_type=INT8U_ATTRIBUTE_TYPE)
    return level_data.data_uint8



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common shell() capability for devices communicating over SSH."""
import re
import time
from typing import Callable, Collection, Optional, Tuple, Union

from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import shell_base
from gazoo_device.switchboard import expect_response

_SSH_CONNECTION_FAILURE_MARKERS = ("Connection to", "Connection reset")

logger = gdm_logger.get_logger()


class ShellSSH(shell_base.ShellBase):
  """Common shell() method implementation for devices communicating over SSH."""

  def __init__(
      self,
      send_and_expect: Callable[..., expect_response.ExpectResponse],
      device_name: str,
      tries: int = shell_base.TRIES,
      timeout: float = shell_base.TIMEOUT,
      failure_markers: Collection[str] = _SSH_CONNECTION_FAILURE_MARKERS
  ) -> None:
    """Initializes an instance of ShellSSH capability.

    Args:
      send_and_expect: send_and_expect method of the device class instance.
      device_name: Name of the device using this capability.
      tries: How many times to try sending the shell command.
      timeout: Shell timeout in seconds.
      failure_markers: Markers indicating SSH connection failure.
    """
    super().__init__(
        send_and_expect=send_and_expect,
        device_name=device_name,
        tries=tries,
        timeout=timeout)
    self._failure_markers = failure_markers

  def shell(
      self,
      command: str,
      command_name: str = "shell",
      timeout: Optional[float] = None,
      port: int = 0,
      include_return_code: bool = False,
      searchwindowsize: int = config.SEARCHWINDOWSIZE
  ) -> Union[str, Tuple[str, int]]:
    """Sends command and returns response and optionally return code.

    If the SSH connection fails, the command is retried (up to a total of
    self._tries attempts).

    Args:
      command: Command to send to the device.
      command_name: Identifier for command.
      timeout: Time in seconds to wait for device to respond.
      port: Which port to send on. Port 0 is typically used for commands.
      include_return_code: Whether to also return the command return code.
      searchwindowsize: Number of the last bytes to look at.

    Raises:
      DeviceError: if communication fails.

    Returns:
      If include_return_code is False, returns the device response to the
        command.
      If include_return_code is True, returns the device response and the
        command return code.
    """
    if timeout is None:
      timeout = self._timeout

    command = command.rstrip()  # Remove trailing newlines.
    command_str = ("echo '{}';".format(command.replace("'", r"'\''")) +
                   command + ";" +
                   "echo Return Code: $?\n")
    command_start_regex = re.escape(command)
    command_end_regex = r"Return Code: (-?\d+)"

    logger.debug("%s sending %r to generate %s in %ds on port %d",
                 self._device_name, command, command_name, timeout, port)

    for attempt in range(self._tries):
      response = self._send_and_expect(
          command_str, [command_start_regex, command_end_regex],
          timeout=timeout,
          port=port,
          searchwindowsize=searchwindowsize,
          expect_type="response",
          mode="sequential")
      if not response.timedout:
        break

      if (any(failure_marker in response.before
              for failure_marker in self._failure_markers) and
          attempt < self._tries - 1):
        logger.warning(
            "{}: SSH connection died with output {}. Trying again.".format(
                self._device_name, response.before))
        # SSH connection died. Retry.
        time.sleep(.1)
      else:
        raise errors.DeviceError("Device {} shell failed for command {!r}. "
                                 "Timed out waiting {}s for response. "
                                 "Shell output: {!r}.".format(
                                     self._device_name, command, timeout,
                                     response.before))

    # Compile shell output and filter results for only our command's output.
    result_list = response.before.splitlines() + response.after.splitlines()
    result_str = response.before + response.after
    pattern = (r"\n" +  # Beginning of a new line.
               command_start_regex + r"\n" +  # The echoed command.
               r"(.*)" +  # The response. May not end in a newline.
               command_end_regex + r"\n")  # The return code.
    match_echo = re.search(pattern, result_str, re.M | re.I | re.DOTALL)
    if match_echo:
      result = match_echo.group(1).strip()
    else:
      result = "\n".join(result_list[1:-1]).strip()

    if include_return_code:
      return_code = int(response.match.group(1))
      return result, return_code
    else:
      return result

  def has_command(self, binary_name: str) -> bool:
    """Returns if binary_name is installed on the device.

    Args:
      binary_name: Name of the executable.
    """
    _, result_code = self.shell(
        f"which {binary_name}\n", include_return_code=True)
    return result_code == 0



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transport properties that can be read or written to for various transports."""

# Common properties
AUTO_REOPEN = "auto_reopen"
OPEN_ON_START = "open_on_start"

# ProcessTransport properties
CLOSE_FDS = "close_fds"

# SerialTransport properties
BAUDRATE = "baudrate"
BYTESIZE = "bytesize"
PARITY = "parity"
STOPBITS = "stopbits"
XONXOFF = "xonxoff"
RTSCTS = "rtscts"
DSRDTR = "dsrdtr"
EXCLUSIVE = "exclusive"
READ_REOPEN = "read_reopen"
USE_HIGH_BAUDRATE_FLOW_CONTROL = "use_high_baudrate_flow_control"

# TcpTransport properties
HOST = "host"
PORT = "port"

# TcpTransport, WebSocketTransport properties
CONNECT_TIMEOUT = "connect_timeout"

# WebSocketTransport properties
WEBSOCKET_URL = "websocket_url"



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the Switchboard capability.

Switchboard is the backbone of all device interaction.
It provides the ability to interact with devices using standardized transport,
button, and expect APIs.

By separating these standardized APIs we can more easily test the logic and
eventually unit test device classes independent of hardware.

Switchboard implementation resides in gazoo_device/switchboard/switchboard.py.
"""
import abc
from typing import List

from gazoo_device import config
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.switchboard import line_identifier

MODE_TYPE_ALL = "all"
MODE_TYPE_ANY = "any"
MODE_TYPE_SEQUENTIAL = "sequential"
VERIFY_METHOD_MD5SUM = "md5sum"


class SwitchboardBase(capability_base.CapabilityBase):
  """Manages device interactions and writes everything to a single file.

  This will spawn and manage 3 or more multiprocess subprocesses:
      Process 0 (Main)
          The primary process where the API/CLI/tests execute from;
          responsible for initializing this module for each device. Performs
          expect and Parser logic and closes subprocesses on shutdown or
          transport errors.

      Process 1..n (Transport)
          One or more subprocesses responsible for performing all device
          transport operations (e.g. open/close and read/write) and
          communicating results to other subprocesses as needed using
          queues.

      Process n+1 (Log writer)
          Performs all log writing operations received from log queue
          shared with Transport subprocess. Log lines are written only for
          completed log lines.

      Process n+2 (Log filter)
          Reads log lines from the log file written by the log writer
          subprocess. Filters each log line read for desired events and
          writes them to an event file. The main process can then use the
          event file to query for relevant events.
  """

  @abc.abstractmethod
  def add_log_note(self, note):
    """Adds given note to device log file.

    Args:
        note (str): to write to the log file
    """

  @abc.abstractmethod
  def add_new_filter(self, filter_path):
    """Adds new log filter at path specified to LogFilterProcess.

    Args:
        filter_path (str): filter file to add

    Raises:
        RuntimeError: if LogFilterProcess is not available or running.
        ValueError: if filter_path doesn't exist
    """

  @abc.abstractmethod
  def click(self, button, duration=0.5, port=0):
    """Press and release the button for the duration and port specified.

    Args:
        button (str): button to press and release
        duration (float): seconds to wait before releasing button
        port (int): which port to click on, 0 or 1.

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, duration, or port values are invalid
    """

  @abc.abstractmethod
  def click_and_expect(self,
                       button,
                       pattern_list,
                       duration=0.5,
                       timeout=30.0,
                       searchwindowsize=config.SEARCHWINDOWSIZE,
                       expect_type="log",
                       port=0,
                       mode="any",
                       raise_for_timeout=False):
    """Press and release button, log lines matching patterns are returned.

    Args:
        button (str): button to press and release
        pattern_list (list): list of regex expressions to look for in the
          lines
        duration (int): seconds to press button before releasing it
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: If buttons are not supported on the device or
                     other arguments are invalid.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before clicking button to catch fast responses.
    """

  @abc.abstractmethod
  def close(self):
    """Shuts down the subprocesses and closes the transports.

    NOTE:
        The current implementation relies on queues being garbage collected.
        Instead of explicitly closing the queues, all queue references MUST
        be deleted to
        release the queues and prevent a memory leak!
    """

  @abc.abstractmethod
  def close_all_transports(self):
    """Leaves the switchboard architecture intact but closes the communication FDs.

    This is used prior to the connections being closed, such as disconnecting an
    ethernet or a serial connection. Only closes the ones open so if
    device.close has already occurred, nothing will be closed.
    """

  @abc.abstractmethod
  def close_transport(self, port=0):
    """Closes the transport specified.

    Args:
        port (int or str): the transport port to close

    Raises:
        DeviceError: If port value is invalid or out of range.
    """

  @abc.abstractmethod
  def do_and_expect(self,
                    func,
                    func_args,
                    func_kwargs,
                    pattern_list,
                    timeout=30.0,
                    searchwindowsize=config.SEARCHWINDOWSIZE,
                    expect_type=line_identifier.LINE_TYPE_LOG,
                    mode=MODE_TYPE_ANY,
                    raise_for_timeout=False):
    """Executes function with given args, blocks until expect matches or timeout occurs.

    Args:
        func (method): name of function to be called
        func_args (list): positional arguments specified to be passed to
          function
        func_kwargs (dict): keyword arguments specified to be passed to
          function
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Raises:
        DeviceError: If func is not callable
                     If other arguments are invalid

    Note:
        Input parameter "func" MUST NOT call "shell" nor another
        "core.xxx_expect" method so as to avoid the nested "flush" problem
        described in 'NEP-2343'.
    """

  @abc.abstractmethod
  def echo_file_to_transport(self,
                             source_file,
                             destination_path,
                             port=0,
                             bytes_per_echo=50):
    r"""Transfers file to transport specified using echo commands.

    Args:
        source_file (path): to the file to transfer
        destination_path (path): to transfer file to on device
        port (int or str): the transport port to open
        bytes_per_echo (int): call to use during file transfer

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port or bytes_per_echo values are invalid or
                     out of range.

    Note:
        The caller is responsible for preparing the device to receive
        multiple echo commands to receive the file and only calling this
        method for devices that support the following commands::

            echo -ne > <destination_path>
            echo -ne "\\x{:02x}" >> <destination_path>
    """

  @abc.abstractmethod
  def ensure_serial_paths_unlocked(self, communication_addresses: List[str]):
    """Ensures serial paths are longer locked by switchboard process after device is closed."""

  @abc.abstractmethod
  def expect(self,
             pattern_list,
             timeout=30.0,
             searchwindowsize=config.SEARCHWINDOWSIZE,
             expect_type=line_identifier.LINE_TYPE_ALL,
             mode=MODE_TYPE_ANY,
             raise_for_timeout=False):
    """Block until a regex pattern is matched or until a timeout time has elapsed.

    Args:
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: if arguments are not valid.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect.
    """

  @abc.abstractmethod
  def get_line_identifier(self):
    """Returns the line identifier currently used by Switchboard."""

  @property
  @abc.abstractmethod
  def number_transports(self) -> int:
    """Returns the number of transport processes used by Switchboard."""

  @abc.abstractmethod
  def open_all_transports(self):
    """Opens the communication FDs, assuming switchboard architecture is intact.

    This is used after a physical connection has been reopened, such as
    reconnecting an ethernet or a serial connection.
    Only opens the ones closed so if device.close has already occurred, nothing
    will be opened.
    """

  @abc.abstractmethod
  def open_transport(self, port=0, timeout=30.0):
    """Opens the transport specified.

    Args:
        port (int or str): the transport port to open
        timeout (float): how long to wait for port to open.

    Raises:
        DeviceError: If port value is invalid or out of range.
    """

  @abc.abstractmethod
  def press(self, button, wait=0.0, port=0):
    """Presses the button for the port specified and waits the time specified.

    Args:
        button (str): button to press
        wait (float): seconds to wait before returning
        port (int): which port to click on, 0 or 1

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, wait, or port values are invalid
    """

  @abc.abstractmethod
  def press_and_expect(self,
                       button,
                       pattern_list,
                       wait=0.0,
                       timeout=30.0,
                       searchwindowsize=config.SEARCHWINDOWSIZE,
                       expect_type="log",
                       port=0,
                       mode="any"):
    """Press button and expect for pattern_list and other arguments provided.

    Args:
        button (str): button to press
        pattern_list (list): list of regex expressions to look for in the
          lines
        wait (float): seconds to wait
        timeout (float): Seconds to look for the patterns
        searchwindowsize (int): Number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): Which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, wait, port, or expect values are invalid

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before pressing button to catch fast responses.
    """

  @abc.abstractmethod
  def release(self, button, port=0):
    """Release the button for the port specified.

    Args:
        button (str): button to release
        port (int): Which port to release button on, 0 or 1

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button or port values are invalid
    """

  @abc.abstractmethod
  def release_and_expect(self,
                         button,
                         pattern_list,
                         timeout=30.0,
                         searchwindowsize=config.SEARCHWINDOWSIZE,
                         expect_type="log",
                         port=0,
                         mode="any"):
    """Release button, matches pattern_list in loglines as specified by expect_type.

    Args:
        button (str): button to release
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, port, or expect values are invalid

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before releasing button to catch fast responses.
    """

  @abc.abstractmethod
  def send(self, command, port=0, slow=False, add_newline=True, newline="\n"):
    """Sends the command to the device on the port (transport) specified.

    Args:
        command (str): to send to the device
        port (int): or transport to send command to
        slow (bool): flag indicating command should be sent byte-by-byte
        add_newline (bool): flag indicating newline should be added to
          command if missing
        newline (str): character to check for and add if missing at the end
          of the command

    Raises:
        DeviceError: if port specified is an invalid value or out of
                     range of the available ports
    """

  @abc.abstractmethod
  def send_and_expect(self,
                      command,
                      pattern_list,
                      timeout=30.0,
                      searchwindowsize=config.SEARCHWINDOWSIZE,
                      expect_type=line_identifier.LINE_TYPE_ALL,
                      mode=MODE_TYPE_ANY,
                      port=0,
                      slow=False,
                      add_newline=True,
                      newline="\n",
                      command_tries=1,
                      raise_for_timeout=False):
    r"""Sends the command and expects on the patterns provided.

    Note: this method does not prepend the command with a wakeup character
    which some devices require. The reason this may be needed is because
    some devices go into a sleep state to save energy and will wakeup on
    receiving the first character sent to it which means the character won't
    get registered into the command buffer. This can be dealt with by
    prepending the command with a nop character that won't affect the
    command being executed in the case that the device has already woken up.
    If there is an issue with this method, try adding "\n" in front of the
    command. E.g. "\nsome_command"

    Args:
        command (str): command to send to the device
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): Seconds to look for the patterns
        searchwindowsize (int): Number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        port (int): Which port to send on, 0 or 1
        slow (bool): flag indicating command should be sent byte-by-byte
        add_newline (bool): flag indicating newline should be added to
          command if missing
        newline (str): character to check for and add if missing at the end
          of the command
        command_tries (int): The number of tries to send the command if it
          times out.
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: if port specified or other expect arguments are
                     invalid, or timed out and raise_for_timeout was True.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an send.
    """

  @abc.abstractmethod
  def set_max_log_size(self, max_log_size):
    """Sets the max_log_size value to the value provided.

    Args:
        max_log_size (int): the max log size to use for log rotation.

    Raises:
        ValueError: if max_log_size is not an integer value
        RuntimeError: if log writer process is not running

    Note:
        A max_log_size of 0 means no log rotation should ever occur.
    """

  @abc.abstractmethod
  def start_new_log(self, log_path):
    """Changes log filter and writer to use a new log path provided.

    Args:
        log_path (str): to log file to switch to

    Raises:
        RuntimeError: if LogWriterProcess is not available or running.
    """

  @abc.abstractmethod
  def transport_jlink_flash(self, image_path, port=0):
    """Sends the flash command to the 'J-Link' transport.

    Args:
        image_path (str): path to the image file to be flashed onto the
          device.
        port (int): the transport port to send the command to.
    """

  @abc.abstractmethod
  def transport_jlink_reset(self, port=0):
    """Sends the reset command to the J - Link transport.

    Args:
        port(int): the transport port to send the command to.
    """

  @abc.abstractmethod
  def transport_serial_set_baudrate(self, new_baudrate, port=0):
    """Sets the serial interface baudrate to a different baudrate.

    Args:
        new_baudrate(int): new baudrate to be set, generally 115200 or
          921600
        port(int or str): the transport port to open

    Raises:
        DeviceError
    """

  @abc.abstractmethod
  def transport_serial_send_xon(self, port=0):
    """Sends the XON control character to the serial interface.

    Args:
        port(int or str): the transport port to open
    """

  @abc.abstractmethod
  def transport_serial_send_xoff(self, port=0):
    """Sends the XOFF control character to the serial interface.

    Args:
        port(int or str): the transport port to open
    """

  @abc.abstractmethod
  def transport_serial_send_break_byte(self, port=0):
    """Sends the break control character to the serial interface (Ctrl + C).

    Args:
        port(int or str): the transport port to open
    """

  @abc.abstractmethod
  def verify_file_on_transport(self,
                               source_file,
                               destination_path,
                               port=0,
                               method=VERIFY_METHOD_MD5SUM):
    """Verifies source file contents matches destination_path on transport using method.

    Args:
        source_file(path): to compare content to on transport
        destination_path(path): to file to verify on transport
        port(int or str): the transport port to open
        method(str): the method to use to verify destination_path

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port or method values are invalid or out of range.

    Returns:
        bool: A boolean status indicating verification was successful.

    Note:
        The caller is responsible for preparing the device to receive one
        of the following verification commands::

            md5sum < destination_path >
    """

  @abc.abstractmethod
  def xmodem_file_to_transport(self, source_file, port=0):
    """Transfers file to transport specified using the XModem protocol.

    Args:
        source_file(path): to the file to transfer
        port(int or str): the transport port to open

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port value provided is invalid or out of range.

    Returns:
        bool: A boolean status indicating xmodem transfer was successful.

    Note:
        The caller is responsible for putting the transport into XModem
        transfer mode before calling this method.
    """

  @abc.abstractmethod
  def add_transport_process(self, transport, **transport_process_kwargs):
    """Add a new transport process to the list of transport processes.

    Args:
        transport(Transport): transport to the device for this process
        **transport_process_kwargs(dict): keyword arguments to the transport
          process
    transport_process_kwargs can be:
        framer(DataFramer): DataFramer derived classes to use to frame
          incoming raw data into raw lines. Defaults to None.
        partial_line_timeout(float): time in seconds to wait before adding
          partial lines to raw_data_queue and log_queue. Defaults to
          transport_process.PARTIAL_LINE_TIMEOUT.
        read_timeout(float): time to wait in seconds for transport reads.
          Defaults to to transport_process._READ_TIMEOUT
        max_read_bytes(int): to attempt to read on each transport read call.
          Defaults to transport_process._MAX_READ_BYTES
        max_write_bytes(int): to attempt to write on each transport write
          call. Defaults to transport_process._MAX_WRITE_BYTES

    Returns:
        int: position of newly added transport process in list of transport
        processes("port")
    """

  @abc.abstractmethod
  def delete_last_transport_process(self):
    """Stops and deletes the last transport process in self._transport_processes.

    Note:
        Just stopping a transport process does not delete it.
        All stopped processes are typically reopened after a device reboot.
        The process must be deleted to ensure it is not reopened after a
        device reboot.

    Since we're using process numbers to identify the transport, deleting any
    transport other than the last one will cause some other transports in the
    transport list to shift their transport number by 1, breaking their usage.
    To prevent this, allow deleting only the last process for now. The proper
    solution would be to use some other form of identification for processes.

    Raises:
        DeviceError: if there's no transport process to delete.
    """

  @abc.abstractmethod
  def start_transport_process(self, process_num):
    """Start the transport process at position process_num in transport list.

    Args:
        process_num(int): position in self._transport_processes list. This
          position is returned by a prior self.add_transport_process() call.

    Raises:
        DeviceError: if process_num has an invalid value.
    """

  @abc.abstractmethod
  def stop_transport_process(self, process_num):
    """Stop the transport process.

    Args:
        process_num(int): number of transport to stop.
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Door Lock endpoint.

This endpoint module corresponds to the
"Door Lock" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Door Lock endpoint houses an instance of Door Lock Matter device type with
the following defined clusters on this endpoint:

The required clusters for this endpoint: Door Lock.
The optional clusters for this endpoint: Groups, Scenes, Alarms, Time,
TimeSync and Poll Control.
"""
from gazoo_device import decorators
from gazoo_device.capabilities.matter_clusters import door_lock_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import door_lock_base


class DoorLockEndpoint(door_lock_base.DoorLockBase):
  """Matter Door Lock endpoint."""

  @decorators.CapabilityDecorator(door_lock_pw_rpc.DoorLockClusterPwRpc)
  def door_lock(self) -> door_lock_pw_rpc.DoorLockClusterPwRpc:
    """Matter Door Lock cluster instance."""
    return self.cluster_lazy_init(door_lock_pw_rpc.DoorLockClusterPwRpc)

  # TODO(b/209362086) Add the below clusters

  # def scenes(self):
  #   """Matter Scenes cluster instance."""
  #   raise NotImplementedError("Scenes cluster isn't implemented.")

  # def groups(self):
  #   """Matter Groups cluster instance."""
  #   raise NotImplementedError("Groups cluster isn't implemented.")

  # def alarms(self):
  #   """Matter Alarms cluster instance."""
  #   raise NotImplementedError("Alarms cluster isn't implemented.")

  # def time(self):
  #   """Matter Time cluster instance."""
  #   raise NotImplementedError("Time cluster isn't implemented.")

  # def time_sync(self):
  #   """Matter TimeSync cluster instance."""
  #   raise NotImplementedError("TimeSync cluster isn't implemented.")

  # def poll_control(self):
  #   """Matter Poll Control cluster instance."""
  #   raise NotImplementedError("Poll Control cluster isn't implemented.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for running Python subprocesses."""

import os
import select
import subprocess
import time
from typing import Sequence, Tuple

from gazoo_device import gdm_logger

logger = gdm_logger.get_logger()

_POLL_DELAY = .2


def run_and_stream_output(cmd_args: Sequence[str],
                          timeout: float) -> Tuple[int, str]:
  """Runs commands and streams the output at the same time.

  Python subprocess cannot capture and stream the output at the same time.
  Here we use Popen to pipe the output and log the output simultaneously.

  Args:
    cmd_args: Command args in a sequence.
    timeout: The timeout for running the commands.

  Returns:
    Tuple of (return code, string output).
    The caller is responsible for verifying that the subprocess invocation is
    successful by checking that the return code is zero and that the output
    doesn't contain any lines indicating a failure.

  Raises:
    subprocess.TimeoutExpired: If the subprocess time outs.
  """
  full_output = []
  logger.info("Calling host command {} with timeout {:0.1f}s. "
              "Streaming command output...".format(cmd_args, timeout))
  start_time = time.time()
  with subprocess.Popen(cmd_args,
                        encoding="utf-8",
                        errors="replace",
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT) as proc:
    os.set_blocking(proc.stdout.fileno(), False)
    end_time = start_time + timeout

    # Keep checking whether the process is terminated by Popen.poll().
    # If the process is still running then we dump and store all the output.
    while proc.poll() is None and time.time() < end_time:
      readable, _, _ = select.select([proc.stdout], [], [], _POLL_DELAY)
      if readable:
        output = readable[0].read()
        logger.info(output.rstrip())
        full_output.append(output)

    # Check the process return_code using Popen.wait(1). If the process is
    # not yet ended then it will raise the TimeoutExpired exception.
    try:
      return_code = proc.wait(timeout=1)
    except subprocess.TimeoutExpired:
      proc.terminate()
      if proc.poll() is None:
        proc.kill()
      raise

  logger.info("Host command {} completed in {:0.1f}s with return code {}."
              .format(cmd_args, time.time() - start_time, return_code))
  return return_code, "".join(full_output)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a serial port transport by wrapping the Pyserial interface."""
import fcntl
import os
import time
import typing

from gazoo_device import gdm_logger
from gazoo_device.switchboard import transport_properties as props
from gazoo_device.switchboard.transports import transport_base
import serial
import six

logger = gdm_logger.get_logger()
DEFAULT_BAUDRATE = 115200
REOPEN_TIMEOUT = 20


class SerialTransport(transport_base.TransportBase):
  """Communication transport for serial port."""

  def __init__(self,
               comms_address,
               baudrate=DEFAULT_BAUDRATE,
               bytesize=serial.EIGHTBITS,
               parity=serial.PARITY_NONE,
               stopbits=serial.STOPBITS_ONE,
               use_high_baudrate_flow_control=False,
               auto_reopen=False,
               open_on_start=True):
    """Initialize the SerialTransport object with the given serial properties.

    Args:
        comms_address (str): the path to the serial device to use
        baudrate (int): the initial baudrate serial property to use
        bytesize (int): the initial byte size serial property to use
        parity (str): the initial parity serial property to use
        stopbits (int): the initial stop bits serial property to use
        use_high_baudrate_flow_control (bool): whether to send control
          characters (XOFF, XON) before & after writing to prevent buffer
          issues at high baud rate (> 115200).
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
    """
    super(SerialTransport, self).__init__(auto_reopen, open_on_start)
    self._max_read_errors = 3
    self._properties.update({
        props.BAUDRATE: baudrate,
        props.BYTESIZE: bytesize,
        props.PARITY: parity,
        props.STOPBITS: stopbits,
        props.XONXOFF: False,
        props.RTSCTS: False,
        props.DSRDTR: False,
        props.EXCLUSIVE: True,
        props.READ_REOPEN: True,
        props.USE_HIGH_BAUDRATE_FLOW_CONTROL: use_high_baudrate_flow_control
    })
    self.comms_address = comms_address
    self._read_errors = 0
    # Windows is not supported due to use of fcntl.
    # Cast to Posix serial so pytype understands this.
    self._serial = typing.cast(serial.serialposix.Serial, serial.Serial())
    self._serial.port = comms_address

  def is_open(self):
    """Returns True if transport is currently open.

    Returns:
        bool: True if transport is currently open, False otherwise.
    """

    return hasattr(self, "_serial") and self._serial.isOpen()

  def _open(self):
    """Opens or reopens the serial port using the current property values.

    Raises:
        SerialException: if serial port path or properties are invalid.
    """

    self._serial.baudrate = self._properties[props.BAUDRATE]
    self._serial.bytesize = self._properties[props.BYTESIZE]
    self._serial.parity = self._properties[props.PARITY]
    self._serial.stopbits = self._properties[props.STOPBITS]
    self._serial.xonxoff = self._properties[props.XONXOFF]
    self._serial.rtscts = self._properties[props.RTSCTS]
    self._serial.dsrdtr = self._properties[props.DSRDTR]
    self._serial.exclusive = self._properties[props.EXCLUSIVE]
    self._serial.timeout = None
    self._serial.write_timeout = None
    self._serial.open()
    # NEP-1852: Prevent inheritance of file descriptors to exec'd child
    # processes
    fd = self._serial.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

  def _close(self):
    """Closes the serial port."""
    # Prevent holding on to the exclusive lock [NEP-2473]
    fcntl.flock(self._serial.fileno(), fcntl.LOCK_UN)
    self._serial.close()

  def _read(self, size=1, timeout=None):
    """Returns bytes read up to max_bytes within timeout in seconds specified.

    Args:
        size (int): maximum number of bytes to read within timeout seconds
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely
                         if timeout=None. Default: None

    Raises:
        SerialException: when read or open raises one

    Returns:
        str: bytes read from transport or None if no bytes were read
    """
    self._serial.timeout = timeout
    try:
      in_waiting = self._serial.in_waiting
      if in_waiting > 3072:
        logger.warning("Serial port input buffer size exceeds 3072 bytes. "
                       "Data will be lost if buffer size exceeds 4095 bytes. \n"
                       "Input buffer size: {}".format(in_waiting))
      result = self._serial.read(size=size)
      if self._read_errors:
        self._read_errors -= 1
    except serial.SerialException as err:
      self._read_errors += 1
      no_data = "read but returned no data" in repr(err)
      read_reopen = self._properties[props.READ_REOPEN]
      if self._read_errors > self._max_read_errors or not no_data or not read_reopen:
        raise
      logger.warn("Recovering from read returned no data, try {}/{}".format(
          self._read_errors, self._max_read_errors))
      # Close and reopen the serial port to fix no data serial port errors
      self.close()
      end_time = time.time() + REOPEN_TIMEOUT
      while not os.path.exists(self._serial.port) and time.time() < end_time:
        time.sleep(0.001)
      self.open()
      # Recursively try again up to max_read_error attempts
      result = self._read(size=size, timeout=timeout)
    return result

  def _write(self, data, timeout=None):
    """Writes data provided within timeout in seconds specified.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely
                         if timeout=None. Default: None

    Raises:
        SerialTimeoutException: in case timeout != None and timeout is
        exceeded
                                when writing to device.

    Returns:
        int: number of bytes written or None if no bytes were written

    Note:
        If the data provided is unicode it will be encoded into UTF-8
        compatible bytes before being written to the serial port.
    """
    if isinstance(data, six.text_type):
      data = data.encode("utf-8", errors="replace")

    # Whether to use flow control to prevent buffer issues at high baudrate
    use_flow_control = (
        self._properties[props.BAUDRATE] > DEFAULT_BAUDRATE and
        self._properties[props.USE_HIGH_BAUDRATE_FLOW_CONTROL])

    if use_flow_control:
      self.send_xoff()

    self._serial.write_timeout = timeout
    bytes_written = self._serial.write(data)

    if use_flow_control:
      self.send_xon()

    return bytes_written

  def set_property(self, key, value):
    """Sets the property specified to the value provided.

    Args:
        key (str): of the property to set
        value (object): of the property to set

    Raises:
        KeyError: when key property doesn't exist
    """
    super(SerialTransport, self).set_property(key, value)

    # Propagate property if device is already open
    if self._serial.isOpen():
      if key == props.BAUDRATE:
        self._serial.baudrate = value
      if key == props.BYTESIZE:
        self._serial.bytesize = value
      if key == props.PARITY:
        self._serial.parity = value
      if key == props.STOPBITS:
        self._serial.stopbits = value
      if key == props.XONXOFF:
        self._serial.xonxoff = value
      if key == props.RTSCTS:
        self._serial.rtscts = value
      if key == props.DSRDTR:
        self._serial.dsrdtr = value
      if key == props.EXCLUSIVE:
        self._serial.exclusive = value

  def flush_buffers(self):
    """Flushes all buffers on the serial device.

    Used during baudrate switching.

    Args: None

    Raises:
        SerialException: when pyserial fails to flush the buffers.

    Returns: None
    """
    self._serial.flush()
    self._serial.reset_input_buffer()
    self._serial.reset_output_buffer()

  def send_xon(self):
    """Sends flow control XON byte.

    Args: None

    Raises:
        SerialException: when pyserial fails to send the XON byte.

    Returns: None
    """
    self._serial.set_input_flow_control(enable=True)

  def send_xoff(self):
    """Sends flow control XOFF byte.

    Args: None

    Raises:
        SerialException: when pyserial fails to send the XOFF byte.

    Returns: None
    """
    self._serial.set_input_flow_control(enable=False)

  def send_break_byte(self):
    """Sends break byte (Ctrl-C).

    Not to be confused by the UART "break condition".

    Args: None

    Raises:
        SerialException: when pyserial fails to send the break byte.

    Returns: None
    """
    self._serial.write(six.b("\x03"))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for GDM logger.

By default, GDM logger handles only a single (main) process.

GDM logger can handle multiple processes: to change to multiprocessing mode,
call switch_to_multiprocess_logging() and call
initialize_child_process_logging(<main process GDM logger queue>) in each child
process that needs to send GDM logs to the main process. The main process's
logger queue (obtained through get_logging_queue()) must be sent as an argument
to child processes.
"""
import atexit
import logging
import logging.handlers
import os
import sys
from typing import List

from gazoo_device import config
from gazoo_device import multiprocess_logging
from gazoo_device.utility import common_utils
from gazoo_device.utility import multiprocessing_utils

# Making this global enables user control of stdout streaming (indirectly)
_stdout_handler = None
# Logging queue to put messages into. Child processes will send their logs to
# this queue.
_logging_queue = None
# Logging thread consumes messages from the logging queue in the main process.
_logging_thread = None

# Log formats for debug
FMT = ('%(asctime)s.%(msecs)03d %(levelname).1s %(process)5d '
       '%(filename)19.19s:%(lineno)d\t%(message)s')
DATEFMT = '%Y%m%d %X'

dirname = os.path.dirname(config.DEFAULT_LOG_FILE)
if not os.path.isdir(dirname):
  os.makedirs(dirname)


def add_handler(handler):
  """Adds a logging handler to the LoggingThread.

  It receives messages sent to GDM Loggers in the main process and its child
  processes.

  Along with remove_handler, allows library users to configure extra
  destinations and formats for log messages emitted from within GDM.

  Args:
      handler (logging.handler): A logging handler
  """
  if _logging_thread:
    _logging_thread.add_handler(handler)
  else:
    get_logger().addHandler(handler)


def flush_queue_messages():
  """Waits until all messages currently in the logger queue are flushed."""
  if _logging_thread:
    _logging_thread.sync()


def get_handlers() -> List[logging.Handler]:
  """Returns the list of active logging handlers."""
  if is_multiprocess_logging_enabled():
    return _logging_thread.handlers
  else:
    return get_logger().handlers


def get_logging_queue():
  """Returns the logging queue used by the GDM logger."""
  return _logging_queue


def get_logger(component_name=None):
  """Returns a Logger that inherits from (or is) the top-level GDM Logger.

  Differs from usual getLogger in that the name given is appended to the name
  of the top-level GDM Logger (e.g. get_logger('component') is equivalent
  to logging.getLogger('gazoo_device_manager.component')), and also in that
  message strings passed to the returned Logger will be combined with the
  given args using .format() rather than %.

  Args:
      component_name (str): name of an GDM component. Nests using '.' char.

  Returns:
      logging.Logger: main logger or sub logger.
  """
  if component_name is not None:
    name = '.'.join(['gazoo_device_manager', component_name])
  else:
    name = 'gazoo_device_manager'

  return logging.getLogger(name)


def initialize_logger():
  """Configures the top-level GDM Logger.

  Configures it to begin logging to stdout and to the default log destination
  (config.DEFAULT_LOG_FILE).
  """
  logger = get_logger()
  logger.setLevel(logging.DEBUG)

  # Ensure no handlers remain during shutdown. GDM logger can log during __del__
  # calls for some objects. This can happen during shutdown. Logging during
  # shutdown can cause crashes in some environments.
  atexit.register(logger.handlers.clear)

  # Configure a handler that writes to GDM logfile
  filepath = config.DEFAULT_LOG_FILE
  args = dict(mode='a', maxBytes=100 * 1024 * 1023, backupCount=5)
  logfile_handler = logging.handlers.RotatingFileHandler(filepath, **args)
  logfile_handler.setLevel(logging.DEBUG)
  logfile_formatter = logging.Formatter(FMT, datefmt=DATEFMT)
  logfile_handler.setFormatter(logfile_formatter)
  logger.addHandler(logfile_handler)

  # Configure a handler that writes INFO logs to stdout
  stdout_handler = logging.StreamHandler(sys.stdout)
  stdout_handler.setLevel(logging.INFO)
  stdout_formatter = logging.Formatter('%(message)s')
  stdout_handler.setFormatter(stdout_formatter)
  logger.addHandler(stdout_handler)

  # Keep global copy of stdout_handler created and added above to allow for
  # changing the log level later for that handler.
  global _stdout_handler
  _stdout_handler = stdout_handler


def is_multiprocess_logging_enabled() -> bool:
  """Returns whether multiprocess logging is enabled."""
  return bool(_logging_queue or _logging_thread)


def switch_to_multiprocess_logging() -> None:
  """Initializes multiprocessing logging in the main process."""
  if is_multiprocess_logging_enabled():
    return

  global _logging_queue
  global _logging_thread
  _logging_queue = multiprocessing_utils.get_context().Queue()
  _logging_queue.cancel_join_thread()
  _logging_thread = multiprocess_logging.LoggingThread(_logging_queue)
  queue_handler = multiprocess_logging.QueueHandler(_logging_queue)
  queue_handler.setLevel(logging.DEBUG)

  logger = get_logger()
  for handler in logger.handlers:  # Transfer log handlers to queue handler.
    _logging_thread.add_handler(handler)
  logger.handlers = [queue_handler]
  if logger.propagate:  # Transfer message propagation to LoggingThread.
    _logging_thread.configure_propagate(logger.propagate, logger.parent)
    logger.propagate = False

  _logging_thread.start()
  atexit.register(common_utils.MethodWeakRef(_logging_thread.stop))


def initialize_child_process_logging(logging_queue):
  """Initializes multiprocessing logging in a child process.

  Used only by GDM child processes (created by Switchboard or parallel_utils),
  which need to use the logging queue sent over by the parent for logging.
  The child's only log handler sends all log messages to the logging queue to be
  logged by the parent.

  Args:
      logging_queue (Queue): multiprocessing queue to put log messages into.
  """
  global _logging_queue
  global _stdout_handler
  _logging_queue = logging_queue
  queue_handler = multiprocess_logging.QueueHandler(logging_queue)
  queue_handler.setLevel(logging.DEBUG)
  logger = get_logger()
  logger.handlers = [queue_handler]
  _stdout_handler = None


def reenable_progress_messages():
  """Reenables streaming GDM Logger messages to stdout."""
  if _stdout_handler:
    add_handler(_stdout_handler)


def remove_handler(handler):
  """Removes the given handler from the GDM Logger's LoggingThread.

  Messages sent to the GDM Logger will no longer be sent to that handler.

  Along with add_handler, allows library users to configure extra destinations
  and formats for log messages emitted from within GDM.

  Args:
      handler (logging.handler): A logging handler that was added earlier
        using add_handler
  """
  if _logging_thread:
    _logging_thread.remove_handler(handler)
  else:
    get_logger().removeHandler(handler)


def set_component_log_level(component_name, log_level):
  """Sets the log level for the named component.

  Args:
      component_name (str): name of an GDM component
      log_level (int): Integer log level, e.g. logging.DEBUG (value is 10)
  """
  logger = get_logger(component_name)
  logger.setLevel(log_level)


def silence_progress_messages():
  """Stops the default behavior of streaming GDM Logger messages to stdout."""
  if _stdout_handler:
    remove_handler(_stdout_handler)


def stream_debug():
  """Sets the log level for GDM Logger stdout streaming to DEBUG."""
  if _stdout_handler:
    fmt = logging.Formatter(FMT, datefmt=DATEFMT)
    _stdout_handler.setFormatter(fmt)
    _stdout_handler.setLevel(logging.DEBUG)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""J-Link transport class."""
import os.path
import sys

from gazoo_device import errors
from gazoo_device.switchboard.transports import transport_base
import intelhex
from pylink import enums
from pylink import jlink

DEFAULT_CHIP_NAME = "NRF52840_XXAA"
_JLINK_NO_DLL_ERROR = "Expected to be given a valid DLL."


class JLinkTransport(transport_base.TransportBase):
  """J-Link transport to a device."""

  def __init__(self,
               comms_address,
               chip_name=DEFAULT_CHIP_NAME,
               speed="auto",
               jlink_interface=enums.JLinkInterfaces.SWD,
               auto_reopen=True,
               open_on_start=True):
    """Initialize the J-Link transport.

    Args:
        comms_address (str): serial number of the J-Link adapter.
        chip_name (str): usually the name of the target chip to connect to.
        speed (object): connection speed, int or str.
                        Valid values: [5-12000], "auto", "adaptive".
        jlink_interface (int): interface to use. See
          pylink.enums.JLinkInterfaces.
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.

    Raises:
        DeviceError: if the J-Link DLL (comes with the J-Link SDK) is
        missing.
        TypeError: J-Link instantiation failed.

    Note: due to license restrictions, J-Link DLL cannot be packaged into GDM.
    """
    super(JLinkTransport, self).__init__(
        auto_reopen=auto_reopen, open_on_start=open_on_start)
    self.comms_address = comms_address
    self._chip_name = chip_name
    self._speed = speed
    self._jlink_interface = jlink_interface

    try:
      self._jlink = jlink.JLink()
    except TypeError as e:
      if _JLINK_NO_DLL_ERROR in str(e):
        raise errors.DependencyUnavailableError(
            "No J-Link DLL found. Install the J-Link SDK from "
            "https://www.segger.com/downloads/jlink/#J-LinkSoftwareAndDocumentationPack. "
            f"Error: {e!r}.")
      else:
        raise

  def close(self):
    """Close the J-Link transport."""
    if hasattr(self, "_jlink") and self._jlink.opened():
      self._stop_rtt_logging()
      self._jlink.close()

  def flash(self, image_path):
    """Flash the hex image onto the device.

    Args:
        image_path (str): path to the image file to be flashed onto the device.
    """
    self._stop_rtt_logging()
    self._jlink.halt()

    image = intelhex.IntelHex(os.path.abspath(image_path))
    for segment_start, segment_end in image.segments():
      segment_size = segment_end - segment_start
      segment = image.tobinarray(start=segment_start, size=segment_size)
      self._jlink.flash_write8(segment_start, segment)

    self.reset()
    self._start_rtt_logging()

  def is_open(self):
    """Returns True if the J-Link library is open and transport is connected to the target.

    Returns:
        bool: True if transport is open, False otherwise.
    """
    return self._jlink.opened() and self._jlink.connected(
    ) and self._jlink.target_connected()

  def reset(self):
    """Reset the board by toggling the reset pin and restarting the CPU."""
    self._jlink.reset()
    self._jlink.restart()

  def _close(self):
    """Override abstract _close hook."""
    pass

  def _open(self):
    """Open the J-Link DLL and connect to the target."""
    self._jlink.open(serial_no=self.comms_address)
    self._jlink.set_tif(self._jlink_interface)
    self._jlink.connect(chip_name=self._chip_name, speed=self._speed)
    self._start_rtt_logging()

  def _read(self, size=1024, timeout=None):
    """Reads up to size bytes from the device RTT buffer.

    Args:
        size (int): maximum number of bytes to read.
        timeout (float): noop.

    Returns:
        bytes: bytes read from transport.
    """
    data = self._jlink.rtt_read(0, size)  # returns a list of int values (bytes)
    return b"".join(value.to_bytes(length=1, byteorder=sys.byteorder)
                    for value in data)

  def _start_rtt_logging(self):
    """Restart RTT logging."""
    self._jlink.rtt_start()

  def _stop_rtt_logging(self):
    """Stop RTT logging."""
    self._jlink.rtt_stop()

  def _write(self, data, timeout=None):
    """Writes the data provided into the device RTT buffer.

    Args:
        data (str): bytes to be written.
        timeout (float): noop.

    Returns:
        int: number of bytes written.
    """
    if isinstance(data, str):
      data = data.encode("utf-8", errors="replace")

    bytes_to_write = list(data)  # convert to list of int values
    return self._jlink.rtt_write(0, bytes_to_write)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for Matter controller capability."""

import abc
from typing import Any, Optional, Sequence

from gazoo_device.capabilities.interfaces import capability_base


class MatterControllerBase(capability_base.CapabilityBase):
  """Abstract base class for Matter controller capability."""

  @property
  @abc.abstractmethod
  def version(self) -> str:
    """Matter SDK version of the controller."""

  @abc.abstractmethod
  def commission(self, node_id: int, setup_code: str,
                 long_discriminator: Optional[int] = None,
                 ssid: Optional[str] = None,
                 password: Optional[str] = None,
                 operational_dataset: Optional[str] = None) -> None:
    """Commissions a device into the controller's fabric.

    Commissioning protocol is based on specified arguments:
      - When operational dataset is provided, pairs the device over ble-thread.
      - When Wi-Fi SSID and password are provided, pairs the device over
        ble-wifi.
      - Otherwise, discover the devices on the network and pairs with the first
        one that matches the setup code and long discriminator if one is
        specified.

    Args:
      node_id: Node ID to assign to the node being commissioned.
      setup_code: Set up PIN code of the remote device.
      long_discriminator: Discriminator of the remote device.
      ssid: Wi-Fi SSID either as a string, or in the form hex:XXXXXXXX where the
        bytes of the SSID are encoded as two-digit hex numbers.
      password: Wi-Fi password, either as a string or as hex data.
      operational_dataset: Thread dataset in base-64. This argument is mutually
        exclusive with ssid and password.
    """

  @abc.abstractmethod
  def decommission(self, node_id: int) -> None:
    """Forgets a commissioned device with the given node id.

    Args:
      node_id: Assigned node id to decommission.
    """

  @abc.abstractmethod
  def read(self, node_id: int, endpoint_id: int, cluster: str,
           attribute: str) -> Any:
    """Reads a cluster's attribute for the given node id and endpoint.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to read the attribute value from.
      attribute: Name of the cluster attribute to read.

    Returns:
      Attribute value of the cluster.
    """

  @abc.abstractmethod
  def write(self, node_id: int, endpoint_id: int, cluster: str, attribute: str,
            value: Any) -> None:
    """Writes a cluster's attribute for the given node id and endpoint.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      value: New attribute value to update the cluster with.
    """

  @abc.abstractmethod
  def send(self, node_id: int, endpoint_id: int, cluster: str, command: str,
           arguments: Sequence[Any]) -> None:
    """Sends a command to a device with the given node id and endpoint.

    Args:
      node_id: Node ID assigned to the commissioned end device.
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to send the command to (e.g. onoff).
      command: Name of the command to send (e.g. toggle).
      arguments: Command arguments.
    """

  @abc.abstractmethod
  def upgrade(self, build_file: str, build_id: str) -> None:
    """Upgrades Matter SDK version of the controller.

    Args:
      build_file: Local path to build file.
      build_id: Commit SHA the controller is built at.
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the fastboot capability."""
import time
from typing import Callable, Dict, Optional

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import fastboot_base
from gazoo_device.utility import adb_utils
from gazoo_device.utility import retry

logger = gdm_logger.get_logger()

_REQUIRED_TIMEOUTS = ("ONLINE",)


class FastbootDefault(fastboot_base.FastbootBase):
  """Fastboot capability."""

  def __init__(self,
               device_name: str,
               communication_address: str,
               fastboot_serial: str,
               timeouts_dict: Dict[str, float],
               add_log_note_fn: Callable[[str], None],
               verify_reboot_fn: Callable[[], None],
               root_device_fn: Callable[[], None]) -> None:
    """Initializes the FastbootDefault capability.

    Args:
      device_name: Name of the device using the capability.
      communication_address: The device ADB serial.
      fastboot_serial: The device fastboot serial.
      timeouts_dict: Dictionary of timeouts.
      add_log_note_fn: Function to add output to log file.
      verify_reboot_fn: Function to wait for the device to boot up.
      root_device_fn: Function to root the device if not rooted.
    """
    super().__init__(device_name=device_name)
    self._communication_address = communication_address
    self._fastboot_serial = fastboot_serial
    self._timeouts_dict = timeouts_dict
    self._add_log_note_fn = add_log_note_fn
    self._verify_reboot_fn = verify_reboot_fn
    self._root_device_fn = root_device_fn

    self.validate_required_keys(_REQUIRED_TIMEOUTS, self._timeouts_dict,
                                "timeouts_dict")

  @decorators.CapabilityLogDecorator(logger)
  def enter(self) -> None:
    """Enters fastboot (bootloader) mode for updating device firmware.

    Raises:
      DeviceError: If unable to boot device into fastboot mode.
    """
    if self._is_in_fastboot_mode():
      return

    output = adb_utils.enter_fastboot(self._communication_address)
    for line in output.splitlines():
      self._add_log_note_fn(line + "\n")

    if not self._is_in_fastboot_mode(timeout=self._timeouts_dict["ONLINE"]):
      raise errors.DeviceError(
          "{} failed to enter fastboot mode in {}s.".format(
              self._device_name, self._timeouts_dict["ONLINE"]))

  @decorators.CapabilityLogDecorator(logger)
  def exit(self) -> None:
    """Exits fastboot (bootloader) mode after updating device firmware.

    Raises:
      DeviceError: If unable to reboot device out of fastboot mode.
    """
    if not adb_utils.is_fastboot_mode(self._fastboot_serial):
      return

    self._add_log_note_fn("GDM triggered reboot via 'fastboot reboot'")
    start_time = time.time()
    output = adb_utils.exit_fastboot(self._fastboot_serial)
    if output:
      for line in output.splitlines():
        self._add_log_note_fn(line + "\n")
      logger.info("%s fastboot reboot took %ds.", self._device_name,
                  time.time() - start_time)
    self._verify_reboot_fn()
    self._root_device_fn()

  def _is_in_fastboot_mode(self, timeout: Optional[float] = None) -> bool:
    """Returns if device in fastboot mode.

    Waits up to timeout time. We try to verify whether the device is in fastboot
    mode in the {timeout} duration, with retry interval 0.5 sec. If timeout is
    None (not provided), there will be no retry.

    Args:
      timeout: Maximum wait time in seconds. If timeout is None, we will not
        retry.

    Returns:
      Whether device is in fastboot mode.
    """
    # No retry when timeout is None. Setting timeout to 0.1 is equivalent to
    # "no retry" since interval is 0.5 sec.
    timeout = timeout or 0.1

    try:
      retry.retry(
          adb_utils.is_fastboot_mode,
          func_args=(self._fastboot_serial,),
          is_successful=bool,
          timeout=timeout,
          interval=0.5)
      logger.info("%s is in fastboot mode.", self._device_name)
      return True
    except errors.CommunicationTimeoutError:
      return False



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ubiquiti UniFi PoE Switch."""
import time

from gazoo_device import config
from gazoo_device import console_config
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import shell_ssh
from gazoo_device.capabilities import switch_power_unifi_switch
from gazoo_device.switchboard import switchboard
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()
CLASS_NAME = "UnifiPoeSwitch"
MARKER = "--- GDM Log Marker ---"
LOGGING_FILE_PATH = "/var/log/messages"
LOG_MARKER_LINE_POS_OR_EMPTY = (
    'grep -n -e "{marker}" {file_path} | tail -n 1 | cut -d '
    '":" -f 1').format(
        marker=MARKER, file_path=LOGGING_FILE_PATH)
LOG_MARKER_LINE_POS = ("line_num=$({cmd}); echo ${{line_num:=1}}".format(
    cmd=LOG_MARKER_LINE_POS_OR_EMPTY))

COMMANDS = {
    "MODEL":
        "mca-cli-op info",
    "FIRMWARE_VERSION":
        "mca-cli-op info",
    "SERIAL_NUMBER":
        "mca-cli-op info",
    "DEVICE_NAME":
        "mca-cli-op info",
    "ADJUST_PORTS_MODE":
        "poe opmode {mode}",
    "GET_PORT_MODE":
        "show poe port 0/{port}",
    "GET_ALL_PORTS_MODE":
        "show poe port all",
    "REBOOT":
        "reboot",
    "NUMBER_OF_PORTS":
        "mca-cli-op info",
    "LOGGING":
        "tail -F -n +$({cmd}) {file_path}".format(
            cmd=LOG_MARKER_LINE_POS, file_path=LOGGING_FILE_PATH),
    "INJECT_LOG_MARKER":
        "sh -c 'echo \"{marker}\" >> {file_path}'".format(
            marker=MARKER, file_path=LOGGING_FILE_PATH),
    "TELNET":
        "echo 'exit' | telnet localhost",
    "MAC_ADDRESS":
        "show hardware"
}

REGEXES = {
    "MODEL_REGEX": r"Model:\s+(\S+)",
    "FIRMWARE_VERSION_REGEX": r"Version:\s+(\S+)",
    "SERIAL_NUMBER_REGEX": r"MAC Address:\s+(\S+)",
    "DEVICE_NAME_REGEX": r"Model:\s+([^-]+)",
    "GET_PORT_MODE_REGEX": r"0/{port}\s+(Auto|Passive24V|Shutdown)",
    "GET_ALL_PORTS_MODE_REGEX": r"(.*)\(Config\)#",
    "NUMBER_OF_PORTS_REGEX": r"Model:\s+\D+(\d+)",
    "MAC_ADDRESS_REGEX": r"Burned In MAC Address\.+ (\S+)"
}

TIMEOUTS = {"SHELL": 10, "SHUTDOWN": 60, "ONLINE": 120}


class UnifiPoeSwitch(auxiliary_device.AuxiliaryDevice):
  """Device class for a Ubiquiti UniFi PoE Switch."""
  COMMUNICATION_TYPE = "SshComms"
  DETECT_MATCH_CRITERIA = {detect_criteria.SshQuery.IS_UNIFI: True}
  DEVICE_TYPE = "unifi_switch"
  _COMMUNICATION_KWARGS = {
      "log_cmd": COMMANDS["LOGGING"],
      "key_info": config.KEYS["unifi_switch_ssh_key"],
      "username": "admin"
  }
  _OWNER_EMAIL = "gdm-authors@google.com"

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)
    self._timeouts.update(TIMEOUTS)

  @decorators.PersistentProperty
  def ip_address(self):
    """IP address."""
    return self.communication_address

  @decorators.DynamicProperty
  def firmware_version(self):
    """Version of UniFi PoE Switch.

    Returns:
        str: UniFi PoE Switch firmware version.
    """
    return self.shell_with_regex(
        command=self.commands["FIRMWARE_VERSION"],
        regex=self._regexes["FIRMWARE_VERSION_REGEX"])

  def get_console_configuration(self) -> console_config.ConsoleConfiguration:
    """Returns the interactive console configuration."""
    return console_config.get_log_response_separate_port_configuration(
        self.switchboard.get_line_identifier())

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [
        self.check_device_connected, self.check_create_switchboard,
        self.check_telnet_connect
    ]

  @decorators.DynamicProperty
  def total_ports(self):
    """Gets the number of ports for the attached unifi_switch.

    Returns:
        int: number of device ports.
    """
    return int(
        self.shell_with_regex(
            command=self.commands["NUMBER_OF_PORTS"],
            regex=self._regexes["NUMBER_OF_PORTS_REGEX"]))

  @decorators.CapabilityDecorator(
      switch_power_unifi_switch.SwitchPowerUnifiSwitch)
  def switch_power(self):
    """switch_power capability to send telnet commands to the device.

    Returns:
        SwitchPowerUnifiSwitch: switch_power_unifi_switch capability.
    """
    return self.lazy_init(
        switch_power_unifi_switch.SwitchPowerUnifiSwitch,
        device_name=self.name,
        command_dict=self.commands,
        regex_dict=self.regexes,
        total_ports=self.total_ports,
        send_and_expect_fn=self.switchboard.send_and_expect,
        send_fn=self.switchboard.send)

  @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
  def shell_capability(self):
    return self.lazy_init(
        shell_ssh.ShellSSH,
        self.switchboard.send_and_expect,
        self.name,
        timeout=self.timeouts["SHELL"],
        tries=2)

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            include_return_code=False):
    """Sends command and returns response and optionally return code.

    Args:
        command (str): Command to send to the device.
        command_name (str): Identifier for command.
        timeout (float): Time in seconds to wait for device to respond.
        port (int): Which port to send on, 0 or 1.
        include_return_code (bool): flag indicating return code should be
          returned.

    Raises:
        DeviceError: if communication fails.

    Note:
        Can try multiple times as connection can sometimes fail.
        See shell_capability init args for setting the number of retry attempts.

    Returns:
        str: If include_return_code is False return the device response to
        the command.
        tuple: If include_return_code is True return the device response and
        return code.
    """
    timeout = timeout or self.timeouts["SHELL"]
    return self.shell_capability.shell(
        command,
        command_name=command_name,
        timeout=timeout,
        port=port,
        include_return_code=include_return_code)

  def shell_with_regex(self,
                       command,
                       regex,
                       regex_group=1,
                       command_name="shell",
                       raise_error=False,
                       tries=1,
                       port=0,
                       timeout=None):
    """Sends a command, searches for a regex in the response, and returns a match group.

    Args:
        command (str): command to issue.
        regex (str): regular expression with one or more capturing groups.
        regex_group (int): number of regex group to return.
        command_name (str): command name to appear in log messages.
        raise_error (bool): whether or not to raise error if unable to find
          a match.
        tries (int): how many times to try executing the command before failing.
        port (int): which port to send the shell command to.
        timeout (float): Time in seconds to wait for device to respond.

    Returns:
        str: value of the capturing group with index 'regex_group' in the match.

    Raises:
        DeviceError: if command execution fails OR couldn't find the requested
                     group in any of the responses.
    """
    return self.command_with_regex(
        command,
        regex,
        self.shell,
        regex_group=regex_group,
        raise_error=raise_error,
        tries=tries,
        command_name=command_name,
        port=port,
        timeout=timeout)

  @decorators.health_check
  def check_telnet_connect(self):
    """Verify it's possible to establish a telnet connection to the CLI.

    Raises:
        DeviceError: if telnet command returns a non zero exit status.
        DeviceNotResponsiveError: if no response to telnet command before
        the timeout.
    """
    try:
      response, return_code = self.shell(
          self.commands["TELNET"], include_return_code=True)

      if return_code != 0:
        raise errors.DeviceError(
            f'{self.name} command {self.commands["TELNET"]}'
            f' returned {return_code} exit status.')

    except Exception as err:
      raise errors.DeviceNotResponsiveError(
          self.name,
          "failed to establish telnet connection",
          timeout=self.timeouts["SHELL"],
          details=str(err))

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of the unifi_poe_switch.

    Returns:
      tuple: (dict, dict) dictionary of persistent attributes,
              dictionary of optional attributes.

    Notes:
        persistent: model, serial_number
    """
    self.props["optional"] = {}
    self.props["persistent_identifiers"]["model"] = self.shell_with_regex(
        command=self.commands["MODEL"], regex=self._regexes["MODEL_REGEX"])
    self.props["persistent_identifiers"][
        "serial_number"] = self.shell_with_regex(
            command=self.commands["SERIAL_NUMBER"],
            regex=self._regexes["SERIAL_NUMBER_REGEX"]).replace(":", "")
    return self.props["persistent_identifiers"], self.props["optional"]

  @classmethod
  def is_connected(cls, device_config):
    """Checks whether or not the unifi_poe_switch is connected.

    Args:
        device_config (dict): contains "persistent" dict

    Returns:
        bool: whether the device responds to a ping.
    """
    return host_utils.is_pingable(
        device_config["persistent"]["console_port_name"])

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait=False, method="shell"):
    """Reboots the device.

    Verifies device fully boots up afterwards.

    Args:
        no_wait (bool): Return before reboot completes. Default: False
        method (str): reboot technique to use.
    """
    self._inject_log_marker()
    self.switchboard.add_log_note("GDM triggered reboot")
    self.shell(command=self.commands["REBOOT"])
    if not no_wait:
      self._verify_reboot()

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Attempts to recover device based on the type of error specified."""
    if isinstance(error, errors.DeviceNotResponsiveError):
      self.reboot()
    else:
      raise error

  def _verify_reboot(self):
    """Verifies reboot actually occurred."""
    self._ensure_device_goes_offline()
    self._ensure_device_is_online()

  @decorators.CapabilityDecorator(switchboard.SwitchboardDefault)
  def switchboard(self):
    """Instance for communicating with the device."""
    switchboard_name = self._get_private_capability_name(
        switchboard.SwitchboardDefault)
    if not hasattr(self, switchboard_name):
      switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
      switchboard_kwargs.update({
          "communication_address": self.communication_address,
          "communication_type": self.COMMUNICATION_TYPE,
          "log_path": self.log_file_name,
          "device_name": self.name,
          "event_parser": None})
      setattr(self, switchboard_name,
              self.get_manager().create_switchboard(**switchboard_kwargs))

    return getattr(self, switchboard_name)

  def _ensure_device_goes_offline(self, timeout=None):
    """Ensure device is no longer pingable over ssh.

    Args:
        timeout (float): Time in seconds to wait for device to respond.

    Raises:
        DeviceError: Device failed to go offline before the timeout
    """
    timeout = timeout or self.timeouts["SHUTDOWN"]
    start_time = time.time()
    max_disconnect_time = start_time + timeout
    count = 0
    while time.time() < max_disconnect_time:
      if not host_utils.is_pingable(self.ip_address):
        count += 1  # Ensure device is really offline not just a blip
      else:
        count = 0
      if count == 2:
        logger.info("{} offline in {}s.".format(self.name,
                                                int(time.time() - start_time)))
        # close ssh transport as the ssh connection is disconnected.
        self.switchboard.close_all_transports()
        time.sleep(5)  # to ensure offline
        return
      time.sleep(.5)

    raise errors.DeviceError("Failed to go offline within {}s.".format(timeout))

  def _inject_log_marker(self):
    """Adds a log marker to /var/log/syslog.

    This is to prevent reading stale logs.

    Note:
        Device logs are read starting with the last log marker (if present).
    """
    self.shell(self.commands["INJECT_LOG_MARKER"])

  def _ensure_device_is_online(self, timeout=None):
    """Ensure device is online and configs are fully loaded.

    Args:
        timeout (float): Time in seconds to wait for device to respond.

    Raises:
        DeviceError: Device failed to come online before the timeout.
    """
    timeout = timeout or self.timeouts["ONLINE"]
    start_time = time.time()
    max_disconnect_time = start_time + timeout
    while time.time() < max_disconnect_time:
      if host_utils.is_pingable(self.ip_address):
        # There's a delay between the device being responsive to ping
        # and being able to open SSH connections
        time.sleep(10)
        self.switchboard.open_all_transports()
        break
      time.sleep(.5)
    else:
      raise errors.DeviceError("{} failed to become pingable in {}s.".format(
          self.name, timeout))

    output = "Device still offline"
    while time.time() < max_disconnect_time:
      try:
        output, return_code = self.shell(
            self.commands["TELNET"], include_return_code=True)
        if return_code == 0:  # command executed
          logger.info("{} online in {}s".format(self.name,
                                                int(time.time() - start_time)))
          return
      except errors.DeviceError:
        logger.debug(
            "{} failed to respond to {!r}.".format(self.name,
                                                   self.commands["TELNET"]),
            exc_info=True)
      time.sleep(.5)
    raise errors.DeviceError(
        "Failed to come online and respond to {!r} in {}s. Response: {}".format(
            self.commands["TELNET"], timeout, output))


deprecation_utils.add_deprecated_attributes(
    UnifiPoeSwitch,
    [("do_and_expect", "switchboard.do_and_expect", True),
     ("expect", "switchboard.expect", True),
     ("send", "switchboard.send", True),
     ("send_and_expect", "switchboard.send_and_expect", True)])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Embedded script interface for powerswitch.

This class defines the API to manage script over powerswitch.
"""
import abc
from typing import Any, Sequence

from gazoo_device.capabilities.interfaces import capability_base


class EmbeddedScriptBase(capability_base.CapabilityBase):
  """Abstract base class defining the API for the embedded script capability."""

  @abc.abstractmethod
  def run(
      self, script_name: str, script_args: Sequence[Any] = ()) -> str:
    """Runs the embedded script over powerswitch.

    Args:
      script_name: Name of the script to run.
      script_args: List of arguments if any consumed by script in order.

    Returns:
      response: Formatted GET/POST HTTP response.
    """

  @abc.abstractmethod
  def terminate(self, thread_id: str = "all") -> str:
    """Terminates the thread executing the script.

    Terminates thread with provided thread_id, if no thread_id provided
    terminates all threads under execution.

    Args:
      thread_id: Id of thread executing the script.
        Defaults to 'all', terminating all threads under execution.

    Returns:
      str: Formatted GET/POST HTTP response.
    """



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gazoo Device Manager (gdm) is a single interface for device interactions.

It manages communication and setup of smart devices to simplify daily tasks
(software updating a unit, getting logs) as well as serving as a device driver
for automated (python) tests.

Areas of focus: ease of use (including ease of installation, setup, updating),
robustness, scalability.

Examples of functionality:
* Detecting and setting up devices when they are attached (create config
  files automatically).
* Setting up shared permissions/file locations so all users can access test
  hardware (if they can log onto the pc).
* Sharing a single device UART for logging and commands, to allow logs to be
  streamed "live" during a test while still sending commands/responses during a
  test.

Command Line Interface (CLI):
There two groups of commands.
"Manager" commands operate on local config files on the pc.
  The command may talk to devices (detect), but they don't change the
  state of the device (i.e. read-only).
  The command may change the config files stored on the pc, to detect
  new devices, add an alias, etc.
  For manager commands, the format is
    "gdm <command> <device> [<additional arguments>]".
    For example: "gdm log cambrionix-1234".
"Device" commands talk to or configure connected devices.
  This includes upgrade, reboot, etc.
  For device commands, the format is
    "gdm issue <device> - <command> [<arguments>]".
  For example: "gdm issue cambrionix-1234 - reboot".

You can get more details on a particular command at the command line with
"gdm man DEVICENAME COMMAND" (does not require a connected device) or
"gdm issue DEVICENAME - COMMAND -- --help" (requires a connected device).
For example: "gdm issue cambrionix-1234 - reboot -- --help"
             "gdm man cambrionix reboot"

Refer to https://github.com/google/gazoo-device for full documentation.
"""
import gc
import logging
import multiprocessing
import signal
import sys

from gazoo_device import _version
from gazoo_device import gazoo_device_controllers
from gazoo_device import gdm_logger
from gazoo_device import manager
from gazoo_device import mobly_controller
from gazoo_device import package_registrar
from gazoo_device.utility import multiprocessing_utils
from gazoo_device.utility import signal_utils

Manager = manager.Manager
register = package_registrar.register
version = _version.version
__version__ = _version.version

# For Mobly integration
MOBLY_CONTROLLER_CONFIG_NAME = "GazooDevice"
create = mobly_controller.create
destroy = mobly_controller.destroy
get_info = mobly_controller.get_info
get_manager = mobly_controller.get_manager

multiprocessing_utils.configure_multiprocessing()

# Defend against inadvertent basicConfig, which adds log noise
logging.getLogger().addHandler(logging.NullHandler())

# Ensure that 'finally' clauses and atexit handlers run when killed by SIGTERM.
signal.signal(signal.SIGTERM, signal_utils.handle_sigterm)


# Set up logger
gdm_logger.initialize_logger()


# Register device classes and capabilities built into GDM
package_registrar.register(gazoo_device_controllers)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of Matter Occupancy Sensing cluster capability.
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import occupancy_base
from gazoo_device.protos import attributes_service_pb2

_OccupancyCluster = matter_enums.OccupancySensingCluster
BITMAP_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_BITMAP8_ATTRIBUTE_TYPE


class OccupancyClusterPwRpc(occupancy_base.OccupancyClusterBase):
  """Matter Occupancy Sensing cluster capability."""

  @decorators.DynamicProperty
  def occupancy(self) -> BITMAP_ATTRIBUTE_TYPE:
    """The Occupancy attribute.

    This attribute is a bitmap where bit 0 specifies the sensed occupancy as
    follows: 1 = occupied, 0 = unoccupied. All other bits are reserved.

    Returns:
      The Occupancy attribute.
    """
    occupancy_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_OccupancyCluster.ID,
        attribute_id=_OccupancyCluster.ATTRIBUTE_OCCUPANCY,
        attribute_type=BITMAP_ATTRIBUTE_TYPE)
    return occupancy_data.data_uint8

  @decorators.DynamicProperty
  def occupancy_sensor_type(self) -> int:
    """The OccupancySensorType attribute.

    This attribute specifies the type of the occupancy sensor:
    0 = PIR; 1 = Ultrasonic; 2 = PIR and ultrasonic; 3 = Physical contact

    Returns:
      The OccupancySensorType attribute.
    """
    sensor_type_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_OccupancyCluster.ID,
        attribute_id=_OccupancyCluster.ATTRIBUTE_OCCUPANCY_SENSOR_TYPE,
        attribute_type=BITMAP_ATTRIBUTE_TYPE)
    return matter_enums.OccupancySensorType(sensor_type_data.data_uint8)

  @decorators.DynamicProperty
  def occupancy_sensor_type_bitmap(self) -> BITMAP_ATTRIBUTE_TYPE:
    """The OccupancySensorTypeBitmap attribute.

    The attribute specifies the types of the occupancy sensor: a "1" in each bit
    position indicates this type is implemented: Bit 0 = PIR; Bit 1 = Ultrasonic
    ; Bit 2 = Physical contact.

    Returns:
      The OccupancySensorTypeBitmap attribute.
    """
    sensor_type_bitmap = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_OccupancyCluster.ID,
        attribute_id=
        _OccupancyCluster.ATTRIBUTE_OCCUPANCY_SENSOR_TYPE_BITMAP,
        attribute_type=BITMAP_ATTRIBUTE_TYPE)
    return sensor_type_bitmap.data_uint8



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""echo flavor of file transfer capability."""
import base64
import os

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import file_transfer_base

logger = gdm_logger.get_logger()


class FileTransferEcho(file_transfer_base.FileTransferBase):
  """echo file transfer capability."""
  _REQUIRED_COMMANDS = ["ECHO_FILE_IN_BASE64", "IS_DIRECTORY"]

  def __init__(self,
               device_name,
               command_dict,
               shell_fn,
               echo_file_to_transport_fn,
               send_tries=1):
    """Initialize echo file transfer capability.

    Args:
        device_name (str): name of the device using the capability.
        command_dict (dict): A dictionary containing the commands used by
          the capability.
        shell_fn (func): The function that can be called to send a shell
          command to the device.
        echo_file_to_transport_fn(func): streams file to device in chunks
          and writes it to source file on device.
        send_tries (int): number of times to try the file transfer process.
    """
    super(FileTransferEcho, self).__init__(device_name=device_name)
    self._command_dict = command_dict
    self._shell_fn = shell_fn
    self._send_tries = send_tries
    self._echo_file_to_transport_fn = echo_file_to_transport_fn

    self.validate_required_keys(self._REQUIRED_COMMANDS, self._command_dict,
                                "command_dict")

  @decorators.CapabilityLogDecorator(logger)
  def recv_file_from_device(self, src, dest="./"):
    """Copies srcs on device to dest on host.

    Args:
        src (str): file path on device to copy to host.
        dest (str): destination path on host computer.

    Raises:
        DeviceError: if destination directory doesn't exist or copy failed.

    Note:
        If no dest is provided, the file will be copied to the current
        working directory on the host computer.
    """
    destination_dir = os.path.dirname(dest)
    if destination_dir != "." and not os.path.exists(destination_dir):
      raise errors.DeviceError(
          "Device {} receive from device failed. "
          "Destination directory {} doesn't appear to exist.".format(
              self._device_name, destination_dir))
    if os.path.isdir(dest):
      dest = os.path.join(dest, os.path.basename(src))
    logger.info("%s receiving file from device. Source: %s Destination: %s",
                self._device_name, src, dest)

    output, ret_code = self._shell_fn(
        self._command_dict["ECHO_FILE_IN_BASE64"].format(src),
        include_return_code=True)
    if ret_code != 0:
      raise errors.DeviceError(
          "Response to base64 encoding: {!r}".format(output))

    output += "===="  # extraneous padding addresses incorrect padding error (Python bug 33770)
    decoded_contents = base64.decodebytes(
        output.encode("utf-8", errors="replace"))

    with open(dest, "wb") as open_file:
      open_file.write(decoded_contents)

    if not os.path.exists(dest):
      raise errors.DeviceError("File {} does not exist".format(dest))

  @decorators.CapabilityLogDecorator(logger)
  def send_file_to_device(self, src, dest):
    """Copies src from host to dest on the device.

    Args:
        src (str): local file path on host computer.
        dest (str): file path on device where the file should be copied to.

    Raises:
        DeviceError: if source file doesn't exist or copy failed.
    """
    if not os.path.exists(src):
      raise errors.DeviceError("Device {} send to device failed. "
                               "Source file {} doesn't appear to exist.".format(
                                   self._device_name, src))

    if self._is_valid_device_directory(dest):
      dest = os.path.join(dest, os.path.basename(src))
    logger.info("%s sending file to device. Source: %s Destination: %s",
                self._device_name, src, dest)

    for try_num in range(self._send_tries):
      try:
        self._echo_file_to_transport_fn(src, dest)
        return
      except errors.DeviceError as err:
        logger.info("{} failed to send to device {!r} - Retrying.".format(
            self._device_name, src))
        if try_num == self._send_tries - 1:
          raise err

  def _is_valid_device_directory(self, directory):
    _, ret_code = self._shell_fn(
        self._command_dict["IS_DIRECTORY"].format(directory),
        include_return_code=True)
    return ret_code == 0



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides classes to ease writing snippets of javascript code in python.

The JsSelector class allows generating selector code easily.
"""


class JsSelector(object):
  """Create javascript code for selecting DOM elements.

  This class generates javascript code that can chain calls on a root element
  ('document' by default) while catering to common use cases such as
  querySelector, and shadowRoot.  Uses a javascript array type as an optional
  type so that you can create a document element selector arbitrarily deep,
  across shadowroots, and with null guards.

  Typical usage:
      RECIPE = (
          JsSelector()
          .QuerySelector('[data-window-id="ui.RECIPE_OVERVIEW"]').NullGuard()
          .ShadowRoot().NullGuard()
          .QuerySelector('[id="some-container"]').NullGuard())
      RECIPE_ELEMENT = RECIPE.Property('someElement')

      def test():
          <device>.devtools.connect().evaluate_javascript(
              RECIPE_ELEMENT.GetOrDefault('0'))
          <device>.devtools.connect().evaluate_javascript(
              '(%s) != null' %s RECIPE.GetOrDefault())
  """

  def __init__(self, javascript='[document]'):
    self._javascript = '%s' % javascript

  def __str__(self):
    """Returns the javascript representation as a description."""
    return self._javascript

  def QuerySelector(self, selector):
    """Returns this selector followed by a call to querySelector."""
    return JsSelector('%s.map(e=>e.querySelector(\'%s\'))' %
                      (self._javascript, selector))

  def NullGuard(self):
    """Returns this selector followed by a null check.

    Generates code that ensures that subsequent javascript methods will not
    be called if the value is null.

    Returns:
        JsSelector: A selector that represents this selector followed by a
        null check.
    """
    return JsSelector('%s.flatMap(e=>e==null?[]:[e])' % (self._javascript))

  def ShadowRoot(self):
    """Returns this selector followed by getting the shadowRoot property."""
    return self.Property('shadowRoot')

  def Property(self, property_name):
    """Returns this selector followed by a call to the given property."""
    return JsSelector('%s.map(e=>e.%s)' % (self._javascript, property_name))

  def Map(self, function):
    """Returns this selector followed by a call to the given function."""
    return JsSelector('%s.map(%s)' % (self._javascript, function))

  def GetOrDefault(self, default='null'):
    """Returns code for this selector's value or the given default.

    The expression will yield either null or the value.

    Args:
        default (str): A javascript expression of the default value. Must
          not be None, but the javascript expression may evaluate to 'null'.

    Returns:
        str: a string with the javascript expression.
    """
    return '[%s].flatMap(a=>a.length==0?[(%s)]:a)[0]' % (self._javascript,
                                                         default)

  def ToJavascript(self):
    """Returns the javascript code for this selector.

    The statement will have a type of array, and the array will be empty
    or contain a single value.

    Returns:
        str: a string with the javascript expression.
    """
    return self._javascript
